import inspect
import re
import typing
import warnings
from collections import deque
from decimal import Decimal
from enum import Enum, EnumMeta
from functools import partial
from typing import (Any, AsyncGenerator, Callable, Dict, Generator, List,
                    Mapping, Optional, Tuple, Type, TypeVar, Union, Iterator)

from ..utils import exceptions as exc
from ..utils.compat import (ForwardRef, Literal, evaluate_forward_ref,
                            get_args, get_origin)
from ..utils.datastructures import unprovided
from ..utils.functional import multi, pop
from ..utils.transform import TypeTransformer
from .options import RuntimeContext

T = typing.TypeVar("T")
OTHER = TypeVar("OTHER")
ORIGIN = TypeVar("ORIGIN")

NUM_TYPES = (int, float, Decimal)
SEQ_TYPES = (list, tuple, set, frozenset, deque, Iterator)
MAP_TYPES = (dict, Mapping)
TYPE_EXACT_TOLERANCE = ({int, float}, {int, Decimal}, (float, Decimal))

OPERATOR_NAMES = {
    "&": "AllOf",
    "|": "AnyOf",
    "^": "OneOf",
    "~": "Not",
}
NONE_ARG_ALLOWED_TYPES = (Callable, Union, Generator, AsyncGenerator)


def resolve_forward_type(t):
    if isinstance(t, ForwardRef):
        if t.__forward_evaluated__:
            return t.__forward_value__, True
        return t, False
    elif isinstance(t, LogicalType):
        res = t.resolve_forward_refs()
        return t, res
    return t, False


def register_forward_ref(
    annotation,
    constraints: dict = None,
    global_vars: Dict[str, Any] = None,
    forward_refs: Dict[str, Tuple[ForwardRef, dict]] = None,
    forward_key: str = None,
    force_clear: bool = False
):

    if not isinstance(annotation, ForwardRef):
        return
    evaluated = None
    if annotation.__forward_evaluated__:
        evaluated = True
        annotation = annotation.__forward_value__
    elif global_vars:
        ref = annotation
        try:
            annotation = evaluate_forward_ref(annotation, global_vars, None)
        except NameError:
            # ignore for now
            pass
        else:
            evaluated = True
            if force_clear:
                ref.__forward_evaluated__ = False
                ref.__forward_value__ = None

    if not evaluated:
        if isinstance(forward_refs, dict):
            # class A:
            #   attr1: 'forward' = Field(gt=1)
            #   attr2: 'forward' = Field(gt=2)
            # we use forward_key (attname) over forward_arg
            forward_refs.setdefault(
                f"${forward_key}" if forward_key else annotation.__forward_arg__,
                # use a $ to differ from forward arg
                (annotation, constraints),
            )
            # still not evaluated
            return annotation
        # raise TypeError(f'{repr(forward_key)}: Unsupported ForwardRef: {annotation}')
    return annotation


class LogicalType(type):  # noqa
    def __instancecheck__(cls, obj):
        if isinstance(obj, LogicalType):
            return super().__instancecheck__(obj)
        if cls.combinator:
            for arg in cls.args:
                if isinstance(arg, type) and isinstance(obj, arg):
                    return True
            return False
        origin = getattr(cls, "__origin__", None)
        if isinstance(origin, type):
            if not isinstance(obj, origin):
                return False
            try:
                cls(obj)
                return True
            except exc.ParseError:
                return False
        return False

    @property
    def args(cls):
        return cls.__dict__.get("__args__", [])

    @property
    def combinator(cls):
        return cls.__dict__.get("__combinator__", None)

    def resolve_origins(cls) -> List[type]:
        if cls.combinator:
            origins = []
            for arg in cls.args:
                if isinstance(arg, LogicalType):
                    for ori in arg.resolve_origins():
                        if ori not in origins:
                            origins.append(ori)
                elif isinstance(arg, type):  # exclude forward ref
                    if arg not in origins:
                        origins.append(arg)
            return origins

        origin = getattr(cls, "__origin__", None)
        if isinstance(origin, LogicalType):
            return origin.resolve_origins()
        elif isinstance(origin, type):
            return [origin]
        return []

    def resolve_combined_origin(cls) -> Optional["LogicalType"]:
        if cls.combinator:
            return cls
        origin = getattr(cls, "__origin__", None)
        if isinstance(origin, LogicalType):
            return origin.resolve_combined_origin()
        return None

    @classmethod
    def _parse_arg(mcs, arg):
        if arg is None:
            return type(None)

        if isinstance(arg, ForwardRef):
            return arg

        if isinstance(arg, mcs):
            return arg

        __origin = get_origin(arg)
        if __origin:
            # like List[str] Literal["value"]
            _new_args = get_args(arg) or ()
        else:
            # only if not origin is detected, we let go the class arg
            # for py>3.8, something like list[int] / dict[str, int] in also a class
            if isinstance(arg, type):
                return arg
            __origin = Literal
            _new_args = (arg,)
        return Rule.annotate(__origin, *_new_args)

    def resolve_forward_refs(cls):
        if not cls.combinator:
            return
        args = []
        resolved = False
        for i, arg in enumerate(cls.args):
            arg, resolved = resolve_forward_type(arg)
            if resolved:
                arg = cls._parse_arg(arg)
            args.append(arg)
        if resolved:
            # only adjust args if resolved
            setattr(cls, "__args__", tuple(args))
        return resolved

    def register_forward_refs(
        cls,
        global_vars: Dict[str, Any] = None,
        forward_refs=None,
        forward_key: str = None,
        force_clear: bool = False
    ):
        if not cls.combinator:
            return
        args = []
        registered = False
        for i, arg in enumerate(cls.args):
            key = f"{forward_key}:{i}" if forward_key else str(i)
            if isinstance(arg, ForwardRef):
                arg = register_forward_ref(
                    annotation=arg,
                    global_vars=global_vars,
                    forward_refs=forward_refs,
                    forward_key=key,
                    force_clear=force_clear,
                )
                arg = cls._parse_arg(arg)
                registered = True
            elif isinstance(arg, LogicalType) and arg.combinator:
                if arg.register_forward_refs(
                    global_vars=global_vars,
                    forward_refs=forward_refs,
                    forward_key=key,
                    force_clear=force_clear
                ):
                    registered = True
            args.append(arg)
        if registered:
            # only adjust args if registered
            setattr(cls, "__args__", tuple(args))
        return registered

    @classmethod
    def combine(mcs, operator: str, *args):
        __args = []
        for arg in args:
            if isinstance(arg, str):
                arg = ForwardRef(arg)

            arg = mcs._parse_arg(arg)

            if arg in __args:
                # avoid duplicate
                continue
            __args.append(arg)

        return mcs(
            OPERATOR_NAMES.get(operator, operator),
            (),
            {"__args__": __args, "__combinator__": operator},  # noqa
        )  # noqa

    def combine_by(cls, comb: str, other, reverse: bool = False):
        if cls.combinator == comb:
            left_parts = cls.args
        else:
            left_parts = (cls,)
        if isinstance(other, LogicalType) and comb == other.combinator:
            right_parts = other.args
        else:
            right_parts = other if isinstance(other, tuple) else (other,)
        args = (*right_parts, *left_parts) if reverse else (*left_parts, *right_parts)
        return cls.combine(comb, *args)

    @classmethod
    def all_of(mcs, *args):
        return mcs.combine("&", *args)

    @classmethod
    def any_of(mcs, *args):
        return mcs.combine("|", *args)

    @classmethod
    def one_of(mcs, *args):
        return mcs.combine("^", *args)

    @classmethod
    def not_of(mcs, value):
        return mcs.combine("~", value)

    def __and__(cls: T, other: OTHER) -> Union[T, OTHER]:
        return cls.combine_by("&", other)

    def __rand__(cls: T, other: OTHER) -> Union[OTHER, T]:
        return cls.combine_by("&", other, reverse=True)

    def __or__(cls: T, other: OTHER) -> Union[T, OTHER]:
        if getattr(other, "__origin__", None) == Union:
            return cls.combine_by("|", other.__args__)
        return cls.combine_by("|", other)

    def __ror__(cls: T, other: OTHER) -> Union[OTHER, T]:
        if getattr(other, "__origin__", None) == Union:
            return cls.combine_by("|", other.__args__, reverse=True)
        return cls.combine_by("|", other, reverse=True)

    def __xor__(cls: T, other: OTHER) -> Union[T, OTHER]:
        return cls.combine_by("^", other)

    def __rxor__(cls: T, other: OTHER) -> Union[OTHER, T]:
        return cls.combine_by("^", other, reverse=True)

    def __invert__(cls):
        if cls.combinator == "~":
            return cls.args[0]
        return cls.combine("~", cls)

    # def __getitem__(cls, item: T) -> typing.Iterable[T]:
    #     return item

    def __repr__(cls):
        def _repr(_arg):
            if isinstance(_arg, LogicalType):
                return repr(_arg)
            return getattr(_arg, "__name__", None) or repr(_arg)

        if not cls.args:
            origin = getattr(cls, "__origin__", None)
            validators = getattr(cls, "__validators__", [])
            constraints = ", ".join(f"{key}={val}" for key, val, c in validators)
            origin_repr = ""
            if origin:
                origin_repr = _repr(origin)
                if constraints:
                    origin_repr += ", "
            return f"{cls.__name__}({origin_repr}{constraints})"

        args_repr = ", ".join([_repr(arg) for arg in cls.args])
        l_par = "(" if cls.combinator else "["
        r_par = ")" if cls.combinator else "]"
        return f"{cls.__name__}{l_par}{args_repr}{r_par}"

    def __str__(cls):
        return repr(cls)

    def parse(cls, value, *args, **kwargs):
        pass

    # compat Python 3.7, not adding the positional only sign
    def logical_parse(cls, value, context: RuntimeContext = None, **kwargs):
        context = context or RuntimeContext()
        # IMPORTANT:
        # we must do clone here (as the parser do make_runtime)
        # to prompt a new RuntimeOptions, to collect the error in this layer

        if cls.combinator == "&":
            for con in cls.args:
                try:
                    # each value transform will pass on to the next condition
                    # like NormalFloat = AllOf(Float, Not(AbnormalFloat))('3.3')
                    value = context.transformer(value, con)
                except Exception as e:
                    context.handle_error(e)
                    break
            return value

        elif cls.combinator == "|":
            # 1. check EXACT identical type
            for con in cls.args:
                if type(value) == con:
                    return value
            # 2. try to transform in strict mode
            # strict_transformer = options.get_transformer(no_explicit_cast=True, no_data_loss=True)

            for con in cls.args:
                with context.enter(cls.combinator) as new_context:
                    try:
                        # error isolation
                        value = new_context.transformer(value, con)
                    except Exception as e:
                        context.collect_tmp_error(e)
                    else:
                        context.clear_tmp_error()
                        break

        elif cls.combinator == "^":
            # 1. check EXACT identical type
            # because args are de-duplicate, so value can only end up one type
            for con in cls.args:
                if type(value) == con:
                    return value

            xor = None

            for con in cls.args:
                with context.enter(cls.combinator) as new_context:
                    try:
                        value = new_context.transformer(value, con)
                        if xor is None:
                            xor = con
                        else:
                            context.handle_error(
                                exc.OneOfViolatedError(
                                    f"More than 1 conditions ({xor}, {con}) is True in XOR conditions"
                                )
                            )
                            xor = None
                            break
                    except Exception as e:
                        context.collect_tmp_error(e)

            if xor is not None:
                # only one condition is satisfied in XOR
                context.clear_tmp_error()

        elif cls.combinator == "~":
            for con in cls.args:
                with context.enter(cls.combinator) as new_context:
                    try:
                        new_context.transformer(value, con)
                        context.handle_error(
                            exc.NegateViolatedError(
                                f"Negate condition: {con} is violated"
                            )
                        )
                    except Exception:  # noqa
                        break
                        # value = cls._get_error_result(e, value, **kwargs)

        context.raise_error()  # raise error if collected
        return value

    def __call__(cls, value, *args, **kwargs):
        if cls.combinator:
            return cls.logical_parse(value, *args, **kwargs)
        return cls.parse(value, *args, **kwargs)


class ConstraintMode:
    mode = None
    support_constraints = None

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.value == other.value
        return False


class Lax(ConstraintMode):
    mode = "lax"


class Constraints:
    TYPE_SPEC_CONSTRAINTS = {
        "max_digits": NUM_TYPES,
        "decimal_places": (float, Decimal),
        "multiple_of": NUM_TYPES,
        "unique_items": SEQ_TYPES,
        # "contains": SEQ_TYPES,
        # "max_contains": SEQ_TYPES,
        # "min_contains": SEQ_TYPES,
        # "dependencies": MAP_TYPES,
    }
    CONSTRAINT_TYPES = {
        "length": (int,),
        "max_length": (int,),
        "min_length": (int,),
        # "contains": (type,),
        # "min_contains": (int,),
        # "max_contains": (int,),
        "decimal_places": (int,),
        "max_digits": (int,),
        "multiple_of": (int, float),
        "unique_items": (bool,),
    }
    # default type is resolved to "string"

    def __init__(self, rule_cls: Type["Rule"]):
        self.rule_cls = rule_cls

    @property
    def origin_type(self):
        return self.rule_cls.__origin__

    @origin_type.setter
    def origin_type(self, t: type):
        if isinstance(t, type):
            self.rule_cls.__origin__ = t
            self.rule_cls.__origin_transformer__ = (
                self.rule_cls.transformer_cls.resolver_transformer(t)
            )

    def valid_length(self, bounds: dict):
        length = bounds.get("length")
        max_length = bounds.get("max_length")
        min_length = bounds.get("min_length")
        if length is not None:
            if not (isinstance(length, int) and length >= 0):
                raise exc.ConfigError(f"Rule length: {length} must be a int >= 0")
            if min_length is not None:
                if min_length > length:
                    raise exc.ConfigError(
                        f"Rule length: {length} and min_length: {min_length} both specified"
                    )
                else:
                    pop(bounds, "min_length")
                    min_length = None
            if max_length is not None:
                if max_length < length:
                    raise exc.ConfigError(
                        f"Rule length: {length} and max_length: {max_length} both specified"
                    )
                else:
                    pop(bounds, "max_length")
                    max_length = None
        else:
            if min_length is not None:
                if not (isinstance(min_length, int) and min_length >= 0):
                    raise exc.ConfigError(
                        f"Rule min_length: {min_length} must be a int >= 0"
                    )

                if not min_length:
                    pop(bounds, "min_length")
                    min_length = None

            if max_length is not None:
                if not (isinstance(max_length, int) and max_length > 0):
                    raise exc.ConfigError(
                        f"Rule max_length: {max_length} must be a int > 0"
                    )
                if min_length is not None:
                    if max_length < min_length:
                        raise exc.ConfigError(
                            f"Rule max_length: {max_length} must >= min_length: {min_length}"
                        )

        if {length, min_length, max_length} != {None}:
            # has length constraints, validate type
            if self.origin_type:
                if not hasattr(self.origin_type, "__len__"):
                    # just warning here, we will coerce to str in runtime
                    if issubclass(self.origin_type, (int, float, Decimal)):
                        warnings.warn(
                            f"Rule specify length constraints for type: {self.origin_type} "
                            f"that does not support length, we recommend to use "
                            f'"max_digits" and "round" for number types'
                        )
                    else:
                        warnings.warn(
                            f"Rule specify length constraints for type: {self.origin_type} "
                            f"that does not support length, value will be convert to str to validate length"
                        )

    def valid_bounds(self, bounds: dict):
        _max = _min = None
        _max_t = _min_t = None

        gt = bounds.get("gt")
        ge = bounds.get("ge")
        lt = bounds.get("lt")
        le = bounds.get("le")

        t = self.origin_type
        if t == bool:
            raise TypeError(f"bool type does not support bounds")

        if gt is not None:
            if not callable(gt):
                _min_t = type(gt)
                _min = gt
            if t and not hasattr(t, "__le__"):
                raise exc.ConfigError(
                    f"Rule: type {t} does not support <gt> constraint for not providing __le__ method"
                )

        if ge is not None:
            if _min is not None:
                raise exc.ConfigError("Rule gt/ge cannot assign together")

            if not callable(ge):
                _min_t = type(ge)
                _min = ge

            if t and not hasattr(t, "__lt__"):
                raise exc.ConfigError(
                    f"Rule: type {t} does not support <ge> constraint for not providing __lt__ method"
                )

        if lt is not None:
            if not callable(lt):
                _max_t = type(lt)
                _max = lt

            if t and not hasattr(t, "__ge__"):
                raise exc.ConfigError(
                    f"Rule: type {t} does not support <lt> constraint for not providing __ge__ method"
                )

        if le is not None:
            if _max is not None:
                raise exc.ConfigError("Rule lt/le cannot assign together")

            if not callable(le):
                _max_t = type(le)
                _max = le

            if t and not hasattr(t, "__gt__"):
                raise exc.ConfigError(
                    f"Rule: type {t} does not support <le> constraint for not providing __gt__ method"
                )

        if _min_t and _max_t:
            if _min_t != _max_t:
                raise exc.ConfigError(
                    f"Rule gt/ge type {_min_t} must equal to lt/le type {_max_t}"
                )

        _t = _max_t or _min_t
        if _t:
            if self.origin_type:
                if (
                    isinstance(self.origin_type, LogicalType)
                    and self.origin_type.combinator
                ):
                    resolved = False
                    for arg in self.origin_type.__args__:
                        if {arg, _t} in TYPE_EXACT_TOLERANCE:
                            resolved = True
                            break
                        if issubclass(arg, _t):
                            resolved = True
                            break
                    if not resolved:
                        raise exc.ConfigError(
                            f"Rule range type {_t} not resolved in type: {self.origin_type}"
                        )

                else:
                    if not issubclass(self.origin_type, _t):
                        if {self.origin_type, _t} in TYPE_EXACT_TOLERANCE:
                            pass
                        else:
                            raise exc.ConfigError(
                                f"Rule range type {_t} must equal to value type {self.origin_type}"
                            )
            else:
                self.origin_type = _t

            if _min is not None and _max is not None:
                if _min >= _max:
                    raise exc.ConfigError(
                        f"Rule lt/le ({repr(_max)}) must > gt/ge ({repr(_min)})"
                    )
                if isinstance(_max, int) and isinstance(_min, int):
                    if gt and lt:
                        if _max - _min < 2:
                            raise exc.ConfigError(
                                f"Rule int lt: {_max} - gt: {_min} must greater or equal than 2"
                            )

    def valid_types(self, bounds: dict):
        for key, val in bounds.items():
            value_types = self.CONSTRAINT_TYPES.get(key)
            origin_types = self.TYPE_SPEC_CONSTRAINTS.get(key)
            if value_types and not isinstance(val, value_types):
                raise exc.ConfigError(
                    f"Constraint: {repr(key)} should be {value_types} object, got {val}"
                )
            if origin_types:
                if self.origin_type:
                    if (
                        isinstance(self.origin_type, LogicalType)
                        and self.origin_type.combinator
                    ):
                        if not any(
                            issubclass(tp, origin_types)
                            for tp in self.origin_type.__args__
                        ):
                            raise exc.ConfigError(
                                f"Constraint: {repr(key)} is only for type: "
                                f"{origin_types}, got {self.origin_type.__args__}"
                            )
                    else:
                        if origin_types == NUM_TYPES and self.origin_type == bool:
                            # bool is subclass of int
                            raise exc.ConfigError(
                                f"Constraint: {repr(key)} is only for type: {origin_types}, got bool"
                            )
                        if not issubclass(self.origin_type, origin_types):
                            raise exc.ConfigError(
                                f"Constraint: {repr(key)} is only for type: "
                                f"{origin_types}, got {self.origin_type}"
                            )
                else:
                    # set the type if missing
                    self.origin_type = origin_types[0]

    def validate_constraints(self, constraints: dict):
        if "const" in constraints:
            const = constraints["const"]
            if self.origin_type:
                if isinstance(self.origin_type, LogicalType):
                    raise exc.ConfigError(
                        f"Rule const: {repr(const)} cannot apply to LogicalType"
                    )
                if not isinstance(const, self.origin_type):
                    if {type(const), self.origin_type} in TYPE_EXACT_TOLERANCE:
                        pass
                    else:
                        raise exc.ConfigError(
                            f"Rule const: {repr(const)} not instance of type: {self.origin_type}"
                        )
            # else:
            #     # transform type before check const
            #     self.origin_type = type(const)
            # we do not force a type to const, it is up to developer to decide
            # whether to transform before const check
            return {"const": const}  # ignore other constraints
        elif "enum" in constraints:
            enum = constraints["enum"]
            if isinstance(enum, EnumMeta):
                member_type = getattr(enum, "_member_type_", None)
                if member_type is object:
                    member_type = None
                if member_type:
                    if self.origin_type:
                        if not issubclass(member_type, self.origin_type):
                            raise exc.ConfigError(
                                f"Rule enum member type: {member_type} is "
                                f"conflict with origin type: {self.origin_type}"
                            )
                    else:
                        self.origin_type = member_type
            elif multi(enum):
                enum = list(enum)
            else:
                raise exc.ConfigError(
                    f"Invalid enum: {enum}, must be a Enum subclass of list/tuple/set"
                )
            return {"enum": enum}  # ignore other constraints

        constraints = {k: v for k, v in constraints.items() if v is not None}

        if "unique_items" in constraints and not constraints["unique_items"]:
            # only True is valid
            pop(constraints, "unique_items")

        if "decimal_places" in constraints:
            decimals = constraints["decimal_places"]
            digits = constraints.get("max_digits")
            if decimals and digits:
                # 0.123
                if digits < decimals:
                    raise exc.ConfigError(
                        f"Rule: constraint max_digits: {digits} must >= {decimals}"
                    )

        # other constraints other that const is considered not-null
        self.valid_types(constraints)
        self.valid_bounds(constraints)
        self.valid_length(constraints)
        return constraints

    def validate_base_constraints(self):
        # TODO
        """
        a: PositiveInt = Field(lt=-5)
        like above is considered invalid, but it will be checked in bounds
        some may hard to check
        a: PositiveInt = Field(gt=-5)
        this will override the base constraints, and violate the base class definition
        we should give a warning or error in that case
        """

    def generate_validators(self) -> List[Tuple[str, Any, Callable]]:
        constraint_mode = {}
        constraints = dict()    # for python >= 3.7, ordered dict is dict
        for key in self.rule_cls.__constraints__:
            if hasattr(self.rule_cls, key) and hasattr(self.__class__, key):
                value = getattr(self.rule_cls, key)
                if unprovided(value):
                    # this is a sign that the subclass cancel a constraint from the base class
                    # and to differ with None (for const constraint, None is still consider a valid param)
                    continue
                if isinstance(value, ConstraintMode):
                    constraint_mode[key] = value.mode
                    value = value.value
                constraints[key] = value

        if not constraints:
            return []

        constraints = self.validate_constraints(constraints)
        validators = []
        for key, val in constraints.items():
            mode = constraint_mode.get(key)
            if mode:
                name = f"{mode}_{key}"
            else:
                name = key
            func = getattr(self.__class__, name, None)
            if not func:
                raise exc.ConfigError(
                    f"{self.__class__}: constraint {repr(name)} not discovered,"
                    f" you can override this class and custom it"
                )
            validators.append((key, val, func))
        return validators

    @classmethod
    def decimal_places(cls, value, d):
        digits, decimals = cls._parse_decimal(value)

        if decimals > d:
            raise ValueError

        if isinstance(value, Decimal):
            # if current decimal is Decimal('1.3') and decimal places is 2
            # we will make it Decimal('1.30') by using round
            return round(value, d)
        return value

    @classmethod
    def lax_decimal_places(cls, value, r):
        return round(value, r)

    @classmethod
    def multiple_of(cls, value, of: int):
        mod = value % of
        if mod:
            raise ValueError
        return value

    @classmethod
    def lax_multiple_of(cls, value, of: int):
        mod = value % of
        if mod:
            return (value // of) * of
        return value

    @classmethod
    def _parse_decimal(cls, value):
        if not isinstance(value, Decimal):
            value = Decimal(str(value))

        digit_tuple, exponent = value.as_tuple()[1:]
        if exponent in {"F", "n", "N"}:
            raise ValueError

        if exponent >= 0:
            digits = len(digit_tuple) + exponent
            decimals = 0
        else:
            if abs(exponent) > len(digit_tuple):
                digits = abs(exponent)
            else:
                digits = len(digit_tuple)
            decimals = abs(exponent)
        return digits, decimals

    @classmethod
    def max_digits(cls, value, max_digits: int):
        digits, decimals = cls._parse_decimal(value)
        if digits > max_digits:
            raise ValueError
        return value

    @classmethod
    def lax_max_digits(cls, value, max_digits: int):
        digits, decimals = cls._parse_decimal(value)
        if digits <= max_digits:
            return value

        delta = digits - max_digits

        # 123.456
        # decimals: 3
        # max_digits: 4
        # delta: 3

        if decimals >= delta:
            return round(value, decimals - delta)
        raise ValueError

    @classmethod
    def const(cls, value, v):
        if value != v:
            raise ValueError
        # 0 False
        # 1 True
        if type(value) != type(v):
            if {type(value), type(v)} in TYPE_EXACT_TOLERANCE:
                pass
            else:
                raise ValueError
        return v

    @classmethod
    def lax_const(cls, value, v):
        return v

    @classmethod
    def enum(cls, value, lst):
        if isinstance(lst, EnumMeta):
            # return the value instead of the enum type
            return lst(value).value

        if isinstance(value, Enum):
            value = value.value

        if value not in lst:
            raise ValueError
        return value

    @classmethod
    def lax_enum(cls, value, lst):
        if isinstance(lst, EnumMeta):
            # return the value instead of the enum type
            return lst(value).value

        if isinstance(value, Enum):
            value = value.value

        if value not in lst:
            return list(lst)[0]
        return value

    @classmethod
    def regex(cls, value, r):
        if not re.fullmatch(r, str(value)):
            raise ValueError
        return value

    @classmethod
    def gt(cls, value, gt):
        if value <= gt:
            raise ValueError
        return value

    @classmethod
    def ge(cls, value, ge):
        if value < ge:
            raise ValueError
        return value

    @classmethod
    def lax_ge(cls, value, ge):
        if value < ge:
            return ge
        return value

    @classmethod
    def lt(cls, value, lt):
        if value >= lt:
            raise ValueError
        return value

    @classmethod
    def le(cls, value, le):
        if value > le:
            raise ValueError
        return value

    @classmethod
    def lax_le(cls, value, le):
        if value > le:
            return le
        return value

    @classmethod
    def length(cls, value, lg):
        v = value
        if not hasattr(value, "__len__"):
            v = str(value)
        if len(v) != lg:
            raise ValueError
        return value

    @classmethod
    def lax_length(cls, value, lg):
        v = value
        converted = False
        if not hasattr(value, "__len__"):
            converted = True
            v = str(value)
        if len(v) != lg:
            if converted or len(v) < lg:
                raise ValueError
            return value[:lg]
        return value

    @classmethod
    def max_length(cls, value, m):
        v = value
        if not hasattr(value, "__len__"):
            v = str(value)
        if len(v) > m:
            raise ValueError
        return value

    @classmethod
    def lax_max_length(cls, value, m):
        v = value
        converted = False
        if not hasattr(value, "__len__"):
            converted = True
            v = str(value)
        if len(v) > m:
            if converted:
                raise ValueError
            return value[:m]
        return value

    @classmethod
    def min_length(cls, value, m):
        v = value
        if not hasattr(value, "__len__"):
            v = str(value)
        if len(v) < m:
            raise ValueError
        return value

    @classmethod
    def unique_items(cls, value: list, u):
        if not u:
            return value
        lst = []
        for val in value:
            if val in lst:
                raise ValueError(f"value is not unique")
            lst.append(val)
        return value

    @classmethod
    def lax_unique_items(cls, value: list, u):
        if not u:
            return value
        lst = []
        for val in value:
            if val in lst:
                continue
            lst.append(val)
        # if not strict, just return a unique version of the input
        return type(value)(lst)  # noqa


class Rule(metaclass=LogicalType):
    transformer_cls = TypeTransformer
    constraints_cls = Constraints
    context_cls = RuntimeContext

    __origin__: type = None
    __applied__: bool = False
    __abstract__: bool = False
    __args__: Tuple[type, ...] = None
    __ellipsis_args__: bool = False
    __arg_transformers__: Tuple[Callable, ...] = None
    __origin_transformer__: Callable = None
    __args_parser__: Callable = None

    __validators__: List[Tuple[str, Any, Callable]] = []
    __constraints__: List[str] = [
        # define the constraints and it's order
        "gt",
        "ge",
        "lt",
        "le",
        "const",
        "enum",
        "regex",
        "decimal_places",
        "multiple_of",
        "max_digits",
        "length",
        "max_length",
        "min_length",
        # "contains",
        # "max_contains",
        # "min_contains",
        "unique_items",
    ]
    __transformer__: Callable

    # flag for document
    primitive: Literal[
        "null", "boolean", "object", "array", "integer", "number", "string"
    ]
    format: Optional[str]
    extra: dict  # additional data to the generated schema document

    # constraints
    gt: Any  # exclusiveMinimum
    ge: Any  # minimum
    lt: Any  # exclusiveMaximum
    le: Any  # maximum
    const: Any
    enum: Union[Enum, list, tuple, set]
    regex: str
    length: int
    max_length: int
    min_length: int

    # number constraints
    decimal_places: int
    multiple_of: int
    max_digits: int

    # array constraints
    unique_items: bool
    # rule-validate constraints
    contains: type = None
    max_contains: int = None
    min_contains: int = None

    # dependencies: Dict[str, Union[List[str], Dict[str, List[str]]]]     # for dict type only
    # this property can be defined in schema.__options__
    # https://json-schema.org/understanding-json-schema/reference/object.html
    # https://json-schema.org/understanding-json-schema/reference/array.html

    # **kwargs: str -> "additionalProperties": {"type": "string"}

    # json dict can only support str key, we can only extend it to finite primitive types
    # Dict[str, int] patternProperties: {".*": {"type": integer}}
    # Dict[int, int] patternProperties: {"[0-9]": {"type": integer}}, propertyNames: {"type": "integer"}
    # Dict[bool, int] patternProperties: {"true|false": {"type": integer}}, propertyNames: {"type": "boolean"}

    def __class_getitem__(cls, item):
        args = ()
        origin = cls.__origin__
        if isinstance(item, tuple):
            args = item
        elif cls.__origin__:
            args = (item,)
        else:
            origin = item
        if args and cls.__args__:
            raise exc.ConfigError(f"{cls}: args: {cls.__args__} already specified")
        return cls.annotate(origin, *args)

    def __init_subclass__(cls, **kwargs):
        # if not cls.__origin__:
        origin = None
        class_getitem = cls.__dict__.get("__class_getitem__")
        for base in cls.__bases__:
            if issubclass(base, Rule):
                if not class_getitem:
                    class_getitem = getattr(base, "__class_getitem__", None)
                continue
            if origin:
                raise TypeError(f"{cls}: Multiple origin types: {origin}, {base}")
            origin = base

        if origin:
            if not cls.__origin__:
                cls.__origin__ = origin
            elif issubclass(origin, cls.__origin__):
                cls.__origin__ = origin
            else:
                raise TypeError(
                    f"{cls}: Invalid origin: {origin} of sub type:"
                    f" not subclass of base type: {cls.__origin__}"
                )

        if cls.__origin__:
            if isinstance(cls.__origin__, ForwardRef):
                if not cls.__origin__.__forward_evaluated__:
                    raise TypeError(
                        f"{cls}: cannot setup with unevaluated ForwardRef as origin"
                    )
                cls.__origin__ = cls.__origin__.__forward_value__

            if not isinstance(cls.__origin__, type):
                raise TypeError(f"Invalid origin: {cls.__origin__}, must be a class")
            cls.__abstract__ = bool(
                getattr(cls.__origin__, "__abstractmethods__", None)
            )

            cls.__origin_transformer__ = cls.transformer_cls.resolver_transformer(
                cls.__origin__
            )
            if not cls.__origin_transformer__:
                warnings.warn(
                    f"{cls}: origin type: {cls.__origin__} got no transformer resolved, "
                    f"will just pass {cls.__origin__}(data) at runtime"
                )

        if cls.__args__:
            if hasattr(cls, "__class_getitem__"):

                def _cannot_getitem(*_args):
                    raise TypeError(
                        f"{cls.__name__}: argument is already set, "
                        f"cannot perform getitem ({_args})"
                    )

                cls.__class_getitem__ = _cannot_getitem

            if not multi(cls.__args__):
                cls.__args__ = (cls.__args__,)  # noqa
            arg_transformers = []
            for arg in cls.__args__:
                if isinstance(arg, ForwardRef):
                    if arg.__forward_evaluated__:
                        arg = arg.__forward_value__
                    else:
                        # we will resolve it later
                        arg_transformers.append(None)
                        continue
                if arg is None:
                    arg_transformers.append(None)
                    if cls.__origin__ and issubclass(
                        cls.__origin__, NONE_ARG_ALLOWED_TYPES
                    ):
                        continue
                    warnings.warn(
                        f"None arg: {arg} detected where origin type: {cls.__origin__} is not in "
                        f"{NONE_ARG_ALLOWED_TYPES}"
                    )
                    continue

                if not isinstance(arg, type):
                    raise TypeError(f"Invalid arg: {arg}, must be a class")
                transformer = cls.transformer_cls.resolver_transformer(arg)
                if not transformer:
                    warnings.warn(
                        f"{cls}: arg type: {arg} got no transformer resolved, "
                        f"will just pass {arg}(data) at runtime"
                    )

                arg_transformers.append(transformer)
            cls.__arg_transformers__ = tuple(arg_transformers)
            cls.__args_parser__ = cls.resolve_args_parser()
            if not cls.__args_parser__:
                warnings.warn(
                    f"{cls}: type: {cls.__origin__} with __args__ cannot resolve an args parser, "
                    f"you should inherit resolve_args_parser and specify yourself"
                )
        else:
            if class_getitem and not cls.__dict__.get("__class_getitem__"):
                # we should do a little hack here
                # because according to MRO
                # class UniqueTuple(tuple, Array):
                # UniqueTuple[str] will end up calling __class_getitem__ of tuple
                # and get a GenericAlias instead of Rule
                func = partial(class_getitem.__func__, cls)
                func.__func__ = class_getitem.__func__
                cls.__class_getitem__ = func
                # if __args__ is present, we do not set class getitem then

        cls.__validators__ = cls.constraints_cls(cls).generate_validators()
        cls._validate_contains()

    @classmethod
    def annotate(
        cls,
        type_=None,
        *args_,
        constraints: Dict[str, Any] = None,
        global_vars: Dict[str, Any] = None,
        forward_refs=None,
        force_clear_refs=False
    ):
        args = []
        ellipsis_args = False

        if type_ == Any:
            if args_:
                warnings.warn(f"Any type cannot specify args: {args_}")
            if constraints:
                warnings.warn(f"Any type cannot specify constraints: {constraints}")
            return Rule

        elif type_ == Literal:
            # special for literal type
            constraints = constraints or {}
            if len(args_) == 1:
                constraints["const"] = args_[0]
            elif len(args_) > 1:
                constraints["enum"] = args_
            else:
                raise exc.ConfigError(f"empty literal")
            type_ = cls.__origin__ or type(args_[0])
        else:
            for arg in args_:
                if isinstance(arg, TypeVar):
                    type_cons: tuple = getattr(arg, "__constraints__", None)
                    if type_cons:
                        arg = LogicalType.any_of(*type_cons)
                    else:
                        # stand for any
                        # TODO: support type var validation
                        arg = Rule
                    args.append(arg)
                    continue

                if arg is ...:
                    if isinstance(type_, type) and not issubclass(type_, tuple):
                        raise exc.ConfigError(
                            f"{cls} args: {args_} with ... only apply to tuple, got {type_}"
                        )
                    ellipsis_args = True
                    continue
                if arg is None:
                    # some origin like Generator / AsyncGenerator / Union / Callable need None arg
                    args.append(arg)
                    continue
                annotation = cls.parse_annotation(
                    arg,
                    global_vars=global_vars,
                    forward_refs=forward_refs,
                    force_clear_refs=force_clear_refs
                )
                # this annotation can be a ForwardRef
                # not with constraints, cause that is applied to upper layer
                if annotation is None:
                    continue
                args.append(annotation)

        # if not args and not constraints:
        #     return type_

        name = cls.__name__
        if type_ == Union:
            type_ = LogicalType.any_of(*args)
            # clear the args for union
            args = []

        if isinstance(type_, LogicalType):
            # if type_.combinator:
            #     # we cannot directly inherit combined rules
            #     # just add a AllOf condition
            #     return type_ & cls.annotate(
            #         cls.__origin__, *args,
            #         constraints=constraints,
            #         forward_refs=forward_refs,
            #         global_vars=global_vars
            #     )
            if issubclass(type_, cls):
                # use the subclass name
                name = type_.__name__

        # if type_ == Union:
        #     any_of = LogicalType.any_of(*args)
        #     if not constraints:
        #         return any_of
        #     return any_of & cls.annotate(constraints=constraints)   # no need to pass globals

        attrs = {}
        if type_:
            attrs.update(__origin__=type_)
        if args:
            attrs.update(__args__=args)
        if ellipsis_args:
            attrs.update(__ellipsis_args__=True)
        if constraints:
            attrs.update(constraints)
        return LogicalType(name, (cls,), attrs)

    @classmethod
    def parse_annotation(
        cls,
        annotation,
        constraints=None,
        global_vars=None,
        forward_refs=None,
        forward_key=None,
        force_clear_refs=False,
    ):
        if isinstance(annotation, str):
            if not annotation:
                raise TypeError(
                    f"{repr(forward_key)}: Empty forward ref string: {annotation}"
                )
            # ForwardRef
            annotation = ForwardRef(annotation)

        if isinstance(annotation, ForwardRef):
            annotation = register_forward_ref(
                annotation=annotation,
                constraints=constraints,
                global_vars=global_vars,
                forward_refs=forward_refs,
                forward_key=forward_key,
                force_clear=force_clear_refs
            )
            if isinstance(annotation, ForwardRef):
                # if annotation still cannot be resolved by global vars
                # directly return
                return annotation

        if annotation is Any:
            return Rule  # use empty rule as any

        # no constraints, we can directly use it
        if isinstance(annotation, LogicalType):
            if annotation.combinator:
                annotation.register_forward_refs(
                    # we don't need to pass constraints here
                    # as those will be combined in below logics
                    global_vars=global_vars,
                    forward_refs=forward_refs,
                    forward_key=forward_key,
                    force_clear=force_clear_refs
                )
            # do not detect origin for Logical types (including Rule)
            origin = None
        else:
            origin = get_origin(annotation)

        if origin:
            # first resolve origin
            # generic types like list[int] with origin is still a type
            args = get_args(annotation) or ()
            constraints = constraints or {}
            return cls.annotate(
                origin,
                *args,
                constraints=constraints,
                forward_refs=forward_refs,
                global_vars=global_vars,
                force_clear_refs=force_clear_refs
            )
        elif annotation:
            if isinstance(annotation, type):
                if constraints:
                    return cls.annotate(
                        annotation,
                        constraints=constraints,
                        forward_refs=forward_refs,
                        global_vars=global_vars,
                        force_clear_refs=force_clear_refs
                    )
                else:
                    # no constraints, we can directly use it
                    return annotation
            raise TypeError(
                f"{repr(forward_key)}: invalid annotation: {annotation}"
            )
        elif constraints:
            return cls.annotate(
                constraints=constraints,
                forward_refs=forward_refs,
                global_vars=global_vars,
                force_clear_refs=force_clear_refs
            )
        return None

    @classmethod
    def check_type(cls, t):
        return True

    @classmethod
    def merge_type(cls, t):
        if not t or not isinstance(t, type):
            return cls
        if not cls.__origin__:
            return t
        if cls.combinator:
            return t & cls
        if cls.__origin__ and issubclass(cls.__origin__, t):
            return cls
        constraints = {}
        for name, val, func in cls.__validators__:
            constraints[name] = getattr(cls, name, val)
            # do not lose the mode info
            # so we get value from cls first

        # try to find a strong constraint
        if 'const' in constraints or 'enum' in constraints:
            return cls

        if isinstance(t, LogicalType):
            if t.combinator:
                return t & cls
            elif issubclass(t, Rule):
                for name, val, func in t.__validators__:
                    constraints[name] = getattr(t, name, val)
                if 'const' in constraints or 'enum' in constraints:
                    return t
                return Rule.annotate(
                    t.__origin__ or cls.__origin__,
                    *(t.__args__ or cls.__args__ or []),
                    constraints=constraints
                )
        elif issubclass(t, Enum):
            # do not need to apply constraint to a strong type
            return t
        args = cls.__args__ or []
        if not args and not constraints:
            return t
        return Rule.annotate(
            t,
            *args,
            constraints=constraints
        )

    @classmethod
    def parse(cls, value, context: RuntimeContext = None):
        # use __options__ instead of options is to identify much clearer with other subclass init kwargs
        context = context or cls.context_cls()
        options = context.options
        # IMPORTANT:
        # we must do clone here (as the parser do make_runtime)
        # to prompt a new RuntimeOptions, to collect the error in this layer
        value = cls.pre_validate(value, context)

        if cls.__origin__:
            # no matter cls.__transformer__ is None or not
            if cls.__applied__ and isinstance(value, cls.__origin__):
                # this is final for @utype.apply
                # since the actual type is hidden, so the value of the "hidden" type
                # is consider passed the type validation
                # [this property is not recommended to inherited by developer]
                return cls.post_validate(value, context)

            try:
                value = context.transformer.apply(
                    value, cls.__origin__, func=cls.__origin_transformer__
                )
            except Exception as e:
                error = exc.ParseError(origin_exc=e)
                # if type cannot convert, the following args and constraints cannot validate
                # can just abort and stop collect errors if it is specified
                context.handle_error(error, force_raise=True)

            if value is None:
                # do not continue if value is None after parse
                # optional / AnyOf(..., None) may happen to this
                # we just parse args and check constraints for not-None value
                return value

        if cls.__args_parser__:
            value = cls.__args_parser__(value, context)

            if not cls.__abstract__ and type(value) != cls.__origin__:
                # for abstract types (like Sequence / Iterable)
                # we just give an instance that satisfy those abstract methods (like a list instance)
                value = cls.__origin__(value)

        if not options.ignore_constraints:
            # if options ignore constraints, we will just do type transform
            # constraints_inst = cls.constraints_cls(cls, options=options)
            for key, constraint, validator in cls.__validators__:
                # constraint = getattr(cls, key)
                try:
                    value = validator(value, constraint)
                except Exception as e:
                    error = (
                        e
                        if isinstance(e, exc.ConstraintError)
                        else exc.ConstraintError(
                            origin_exc=e, constraint=key, constraint_value=constraint
                        )
                    )
                    # if validator already throw a constraint error
                    # may an inner constraint (like max_contains in contains) is violated
                    context.handle_error(error)

            if cls.contains:
                value = cls._parse_contains(value, context=context)

        context.raise_error()
        # raise error if collected
        # and leave the error the upper layer to collect
        return cls.post_validate(value, context)

    @classmethod
    def pre_validate(cls, value, context: RuntimeContext = None):
        # meant to be inherited to define user-specific logic before type transform
        return value

    @classmethod
    def post_validate(cls, value, context: RuntimeContext = None):
        # meant to be inherited to define user-specific logic after the transform and constraints validation
        return value

    def __init__(self, value):
        # just to let the IDE know that this class can be init with single param
        # this function will not get called
        self._value = self.parse(value)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self._value)})"

    def __str__(self):
        return f"{self.__class__.__name__}({repr(self._value)})"

    @classmethod
    def _validate_contains(cls):
        # validate max_contains and min_contains as well
        contains = cls.contains
        min_contains = cls.min_contains
        max_contains = cls.max_contains

        if max_contains or min_contains:
            if not contains:
                raise exc.ConfigError(
                    f"Rule with max_contains/min_contains must set <contains> constraint"
                )

            if max_contains is not None and min_contains is not None:
                if max_contains < min_contains:
                    raise exc.ConfigError(
                        f"Rule with max_contains: {max_contains} is little than min_contains"
                    )

        elif not contains:
            return

        from collections.abc import Iterable

        if isinstance(cls.__origin__, type) and not issubclass(
            cls.__origin__, Iterable
        ):
            raise exc.ConfigError(
                f"Rule: constraint: contains is only for Iterable, got {cls.__origin__}"
            )

    @classmethod
    def _parse_contains(cls, value, context: RuntimeContext):
        # validate max_contains and min_contains as well
        if not cls.contains:
            return value

        contains = 0
        for i, item in enumerate(value):
            with context.enter(route=i) as item_context:
                try:
                    item_context.transformer(item, cls.contains)
                except (TypeError, ValueError):
                    pass
                else:
                    contains += 1

        if not contains:
            context.handle_error(
                exc.ConstraintError(
                    f"{cls.contains} not contained in value",
                    constraint="contains",
                    constraint_value=cls.contains,
                )
            )
        elif cls.min_contains and contains < cls.min_contains:
            context.handle_error(
                exc.ConstraintError(
                    f"value contains {contains} of {cls.contains}, which is lower than min_contains",
                    constraint="min_contains",
                    constraint_value=cls.min_contains,
                )
            )
        elif cls.max_contains and contains > cls.max_contains:
            context.handle_error(
                exc.ConstraintError(
                    f"value contains {contains} of {cls.contains}, which is bigger than max_contains",
                    constraint="max_contains",
                    constraint_value=cls.max_contains,
                )
            )
        return value

    @classmethod
    def resolve_forward_refs(cls):
        # an override version of LogicalType.resolve_forward_refs
        if not cls.__args__:
            return False
        args = []
        arg_transformers = []
        resolved = False
        for arg, trans in zip(cls.__args__, cls.__arg_transformers__):
            if isinstance(arg, LogicalType):
                # including the Rule class and LogicalType with combinator
                if arg.resolve_forward_refs():
                    resolved = True
            elif isinstance(arg, ForwardRef):
                if arg.__forward_evaluated__:
                    arg = arg.__forward_value__
                    resolved = True
                    transformer = cls.transformer_cls.resolver_transformer(arg)
                    if not transformer:
                        warnings.warn(
                            f"{cls}: arg type: {arg} got no transformer resolved, "
                            f"will just pass {arg}(data) at runtime"
                        )
                    trans = transformer or trans
            args.append(arg)
            arg_transformers.append(trans)
        if resolved:
            cls.__args__ = tuple(args)
            cls.__arg_transformers__ = tuple(arg_transformers)
        return resolved

    @classmethod
    def resolve_args_parser(cls):
        if not cls.__origin__ or not cls.__args__:
            return None
        if issubclass(cls.__origin__, MAP_TYPES):
            return cls._parse_map_args
        elif issubclass(cls.__origin__, SEQ_TYPES):
            if issubclass(cls.__origin__, tuple) and not cls.__ellipsis_args__:
                return cls._parse_tuple_args
            return cls._parse_seq_args
        elif cls.__origin__ == type:
            return cls._parse_type_arg
        return None

    @classmethod
    def _parse_tuple_args(cls, value: tuple, context: RuntimeContext):
        result = []
        options = context.options

        if options.no_data_loss and len(value) > len(cls.__args__):
            for item in range(len(cls.__args__), len(value)):
                context.handle_error(exc.TupleExceedError(item=item, value=value[item]))

        for i, (arg, func) in enumerate(zip(cls.__args__, cls.__arg_transformers__)):
            if i >= len(value):
                context.handle_error(
                    exc.AbsenceError(
                        f"prefixItems required prefix: [{i}] not provided", item=i
                    )
                )

            with context.enter(route=i) as arg_context:
                try:
                    result.append(
                        arg_context.transformer.apply(value[i], arg, func=func)
                    )
                except Exception as e:
                    error = exc.ParseError(
                        item=i, value=value[i], type=arg, origin_exc=e
                    )
                    if options.invalid_items == options.PRESERVE:
                        context.collect_waring(error.formatted_message)
                        result.append(value[i])
                        continue
                    context.handle_error(error)

        return cls.__origin__(result)

    @classmethod
    def _parse_seq_args(cls, value: Union[list, set], context: RuntimeContext):
        result = []
        arg_type = cls.__args__[0]
        arg_transformer = cls.__arg_transformers__[0]
        options = context.options

        for i, item in enumerate(value):
            with context.enter(route=i) as arg_context:
                try:
                    result.append(
                        arg_context.transformer.apply(
                            item, arg_type, func=arg_transformer
                        )
                    )
                except Exception as e:
                    error = exc.ParseError(
                        item=i, value=value[i], type=arg_type, origin_exc=e
                    )
                    if options.invalid_items == options.EXCLUDE:
                        context.collect_waring(error.formatted_message)
                        continue
                    if options.invalid_items == options.PRESERVE:
                        context.collect_waring(error.formatted_message)
                        result.append(item)
                        continue
                    context.handle_error(error)
        return result

    @classmethod
    def _parse_map_args(cls, value: dict, context: RuntimeContext):
        result = {}
        if not cls.__args__:
            return value

        key_type = cls.__args__[0]
        key_transformer = cls.__arg_transformers__[0]
        value_type = None
        value_transformer = None
        if len(cls.__args__) > 1:
            value_type = cls.__args__[1]
            value_transformer = cls.__arg_transformers__[1]

        options = context.options

        for _key, _val in value.items():
            with context.enter(route=f"{_key}<key>") as key_context:
                try:
                    key = key_context.transformer.apply(
                        _key, key_type, func=key_transformer
                    )
                except Exception as e:
                    error = exc.ParseError(
                        item=f"{_key}<key>", value=_key, type=key_type, origin_exc=e
                    )
                    if options.invalid_keys == options.EXCLUDE:
                        context.collect_waring(error.formatted_message)
                        continue
                    elif options.invalid_keys == options.PRESERVE:
                        key = _key
                        context.collect_waring(error.formatted_message)
                    else:
                        context.handle_error(error)
                        continue

            if value_type:
                with context.enter(route=key) as value_context:
                    try:
                        val = value_context.transformer.apply(
                            _val, value_type, func=value_transformer
                        )
                    except Exception as e:
                        error = exc.ParseError(
                            item=key, value=_val, type=value_type, origin_exc=e
                        )
                        if options.invalid_values == options.EXCLUDE:
                            context.collect_waring(error.formatted_message)
                            continue
                        elif options.invalid_values == options.PRESERVE:
                            context.collect_waring(error.formatted_message)
                            val = _val
                        else:
                            context.handle_error(error)
                            continue
            else:
                val = _val
            result[key] = val
        return result

    @classmethod
    def _parse_type_arg(cls, value, context: RuntimeContext):
        type_cls = cls.__args__[0]
        if issubclass(value, type_cls):
            return value
        context.handle_error(exc.InvalidSubclass(
            type=type_cls, value=value
        ))
        return value


if isinstance(Callable, type):
    @TypeTransformer.registry.register(Callable, allow_subclasses=False)
    def transform_callable(transformer: TypeTransformer, value, t):
        if not callable(value):
            raise TypeError
        return value

if isinstance(Any, type):
    @TypeTransformer.registry.register(Any, allow_subclasses=False)
    def transform_callable(transformer: TypeTransformer, value, t):
        # accept any value
        return value


@TypeTransformer.registry.register(metaclass=LogicalType)
def transform_rule(transformer: TypeTransformer, value, t: LogicalType):
    return t(value, context=transformer.context)



from typing import Optional

LOCALS_NAME = "<locals>"


def multi(f):
    return isinstance(
        f, (list, set, frozenset, tuple, type({}.values()), type({}.keys()))
    )


def pop(data, key, default=None):
    if isinstance(data, dict):
        return data.pop(key) if key in data else default
    elif isinstance(data, list):
        return data.pop(key) if key < len(data) else default
    return default


def copy_value(data):
    """
    return a new value identical to default , but different in memory,
    to avoid multiple initialize to modify the same default data
    """
    if multi(data):
        return type(data)([copy_value(d) for d in data])
    elif isinstance(data, dict):
        return {k: copy_value(v) for k, v in data.items()}
    return data


def get_name(func) -> Optional[str]:
    if isinstance(func, str):
        return func
    if isinstance(func, property):
        func = func.fget
    from functools import partial

    if isinstance(func, partial):
        if hasattr(func, "__name__"):
            return func.__name__
        func = func.func
    if hasattr(func, "__name__"):
        return func.__name__
    return None


def get_obj_name(obj) -> str:
    name = getattr(
        obj, "__qualname__", getattr(obj, "__name__", None)
    ) or str(obj)
    if LOCALS_NAME in name:
        name = str(name.split(LOCALS_NAME)[-1]).strip(".")
    return name


def is_local_var(obj):
    name = getattr(
        obj, "__qualname__", getattr(obj, "__name__", None)
    ) or ''
    return not name or LOCALS_NAME in name



import decimal
import io
import json
import re
from collections import deque
from collections.abc import Iterable, Iterator, Mapping, Sequence
from datetime import date, datetime, time, timedelta, timezone
from decimal import Decimal
from enum import Enum
from typing import TYPE_CHECKING, Callable, Optional, Type, TypeVar
from uuid import UUID

from .compat import ForwardRef
from .exceptions import TypeMismatchError
from .functional import multi
from .base import TypeRegistry

if TYPE_CHECKING:
    from ..parser.options import RuntimeContext

T = TypeVar("T")


class DateFormat:
    DATETIME = "%Y-%m-%d %H:%M:%S"
    DATETIME_DF = "%Y-%m-%d %H:%M:%S.%f"
    DATETIME_F = "%Y-%m-%d %H:%M:%S %f"
    DATETIME_P = "%Y-%m-%d %I:%M:%S %p"
    DATETIME_T = "%Y-%m-%dT%H:%M:%S"
    # DATETIME_TZ = "%Y-%m-%dT%H:%M:%SZ"
    # DATETIME_TFZ = "%Y-%m-%dT%H:%M:%S.%fZ"
    DATETIME_TF = "%Y-%m-%dT%H:%M:%S.%f"
    # DATETIME_ISO = "%Y-%m-%dT%H:%M:%S.%fTZD"
    DATETIME_HTTP = "%a, %d %b %Y %H:%M:%S"
    # DATETIME_GMT = "%a, %d %b %Y %H:%M:%S GMT"
    DATETIME_PS = "%a %b %d %H:%M:%S %Y"
    DATETIME_GMT = "%b %d %H:%M:%S %Y"
    DATE = "%Y-%m-%d"
    DATE_HM = "%Y-%m-%d %H:%M"
    # TIME = '%H:%M:%S'


class TypeTransformer:
    registry = TypeRegistry('transformer', shortcut='__transformer__', cache=True)
    
    # ----- preferences
    # can be override
    DATETIME_FORMATS = [
        DateFormat.DATETIME,
        DateFormat.DATETIME_DF,
        DateFormat.DATETIME_F,
        DateFormat.DATETIME_P,
        DateFormat.DATETIME_T,
        DateFormat.DATETIME_TF,
        DateFormat.DATETIME_HTTP,
        DateFormat.DATETIME_PS,
        DateFormat.DATETIME_GMT,
        DateFormat.DATE,
        DateFormat.DATE_HM
    ]
    EPOCH = datetime(1970, 1, 1)
    MS_WATERSHED = int(2e10)
    ARRAY_SEPARATORS = (",", ";")
    NULL_VALUES = ("null", "none", "nil")
    FALSE_VALUES = ("0", "false", "no", "off", "f")
    TRUE_VALUES = ("1", "true", "yes", "on", "t", "y")
    STRUCTURE_BRACKET = [
        "{}",
        "[]",
        "()",
    ]
    DURATION_REGS = [
        re.compile(
            r"^"
            r"(?:(?P<days>-?\d+) (days?, )?)?"
            r"((?:(?P<hours>-?\d+):)(?=\d+:\d+))?"
            r"(?:(?P<minutes>-?\d+):)?"
            r"(?P<seconds>-?\d+)"
            r"(?:\.(?P<microseconds>\d{1,6})\d{0,6})?"
            r"$"
        ),  # Support the sections of ISO 8601 date representation that are accepted by timedelta
        re.compile(
            r"^(?P<sign>[-+]?)"
            r"P"
            r"(?:(?P<days>\d+(.\d+)?)D)?"
            r"(?:T"
            r"(?:(?P<hours>\d+(.\d+)?)H)?"
            r"(?:(?P<minutes>\d+(.\d+)?)M)?"
            r"(?:(?P<seconds>\d+(.\d+)?)S)?"
            r")?"
            r"$"
        ),
    ]

    def __init__(
        self,
        context: "RuntimeContext",
        no_explicit_cast: bool = None,
        no_data_loss: bool = None,
        unresolved_types: str = None,
    ):
        self.context = context
        self.no_explicit_cast = (
            no_explicit_cast
            if no_explicit_cast is not None
            else context.options.no_explicit_cast
        )
        self.no_data_loss = (
            no_data_loss if no_data_loss is not None else context.options.no_data_loss
        )
        self.unresolved_types = (
            unresolved_types
            if unresolved_types is not None
            else context.options.unresolved_types
        )

    @property
    def options(self):
        return self.context.options

    def __repr__(self):
        return f"{self.__class__.__name__}(no_explicit_cast={self.no_explicit_cast}, no_data_loss={self.no_data_loss})"

    @classmethod
    def resolver_transformer(cls, t: type) -> Optional[Callable]:
        return cls.registry.resolve(t)

    def _attempt_from(self, value):
        if self.no_explicit_cast:
            # does not allow
            return value
        if multi(value) and value:
            if self.no_data_loss and len(value) > 1:
                # like convert ["a"] to "a"
                # whether a "data loss" happen in here is leave to future discussion
                # we apply this for now because some data structure will make value a
                # list with a single item
                # * querystring -> query dict
                # * multipart/form-data -> files
                raise TypeError
            return list(value)[0]
        if isinstance(value, Enum):
            return value.value
        return value

    def _from_byte_like(self, data):
        if isinstance(data, (bytes, bytearray, memoryview)):
            if isinstance(data, memoryview):
                data = bytes(data)
            return data.decode(errors="strict" if self.no_data_loss else "ignore")
        return data

    def _attempt_from_number(self, data):
        data = self._from_byte_like(self._attempt_from(data))
        if isinstance(data, datetime):
            return data.timestamp()
        elif isinstance(data, timedelta):
            return data.total_seconds()
        elif isinstance(data, complex) and not self.no_data_loss:
            if not data.imag:
                return data.real
        elif not data:
            # convert ''. None and others to 0
            # might be a questionable feature?
            return 0
        return data

    # ---------------------
    # 2 approach to do first check
    # 1. all the subclasses instance can directly pass through
    # if isinstance(data, t):
    #     # we will follow the type
    #     return data
    #
    # 2. all the subclasses of the hooked types can pass through (by re-initialize)
    # this should guarantee that the type __init__ take only 1 argument
    # if isinstance(data, *_type_classes):
    #     return t(data)

    @registry.register(type(None), allow_subclasses=False)
    def to_null(self, data, t=type(None)) -> None:
        if data is None:
            return None
        if self.no_explicit_cast:
            raise TypeError
        if isinstance(data, str):
            if data.lower() in self.NULL_VALUES:
                return None
        raise TypeError

    # register this metaclass earlier, because str/list/... is all subclass of Iterable
    # this is just a FALLBACK for the iterable/sequence that does not get resolved
    @registry.register(Sequence, Iterable, Iterator)
    def to_iter_types(self, data, t):
        if isinstance(data, t):
            # we will follow the type
            return data
        value = self.to_array_types(data, list)
        if not getattr(t, "__abstractmethods__", None):
            return t(value)
        # if type is still abstracted, just returning the list result
        return value

    @registry.register(Mapping)
    def to_mapping(self, data, t):
        if isinstance(data, t):
            # we will follow the type
            return data
        value = self.to_dict(data, dict)
        if not getattr(t, "__abstractmethods__", None):
            return t(value)
        return value

    @registry.register(str)
    def to_str(self, data, t: Type[str] = str) -> str:
        if isinstance(data, str):
            return t(data)
        data = self._from_byte_like(self._attempt_from(data))
        if self.no_explicit_cast and not isinstance(data, str):
            raise TypeError
        return t(data)

    @registry.register(bytes, bytearray, memoryview)
    def to_bytes(self, data, t: Type[bytes] = bytes):
        data = self._attempt_from(data)

        if isinstance(data, (bytes, bytearray, memoryview)):
            return t(data)

        if isinstance(data, str):
            return t(data.encode())

        if self.no_explicit_cast:
            # bytes can convert all the types (from str)
            # if the value is not bytes (need to transform) in strict mode, throw an error directly
            raise TypeError

        return t(str(data).encode())

    @registry.register(list, tuple, set, frozenset, deque)
    def to_array_types(self, data, t=list):
        if isinstance(data, t):
            return data

        if multi(data):  # same group
            return t(data)

        if self.no_explicit_cast:
            raise TypeError

        data = self._from_byte_like(data)

        if isinstance(data, str):
            data = data.strip()
            # guess string form array like
            # [1, 2, 3]
            # [{"a": b}]
            # {1, 2, 3}
            # a,b,c
            if any(
                data.startswith(v[0]) and data.endswith(v[1])
                for v in self.STRUCTURE_BRACKET
            ):
                try:
                    data = json.loads(data)
                except json.JSONDecodeError:
                    import ast

                    data = ast.literal_eval(data)
                    if multi(data):
                        return t(data)
                else:
                    if isinstance(data, list):
                        return t(data)
            else:
                for sep in self.ARRAY_SEPARATORS:
                    if sep in data:
                        return t(v.strip() for v in data.split(sep))

        # if data is None:
        #     return t()

        if isinstance(data, dict):
            if issubclass(t, set):
                if self.no_data_loss:
                    raise TypeError
                # set cannot pack un-hashable dict item
                return t(data)
            if not data:
                # {} -> [], as for no data loss
                return t()
        # instead of tear the item apart
        return t([data])
        # return t(data)

    @registry.register(dict)
    def to_dict(self, data, t=dict) -> dict:
        if isinstance(data, t):
            return data
        if isinstance(data, Mapping):
            return t(data)

        if self.no_explicit_cast:
            raise TypeError

        if self.no_data_loss:
            if isinstance(data, (list, set, tuple)):
                result = {}
                try:
                    for item in data:
                        # dict([{'a': 1, 'b': 2}]) == {'a': 'b'}
                        # is consider a data loss
                        if isinstance(item, (dict, Mapping)):
                            raise TypeError
                        key, val = item
                        result[key] = val
                    return t(result)
                except (TypeError, ValueError):
                    pass
        else:
            try:
                # try for iterable of key, value pairs
                # but data loss may happen in this case
                # like dict([{"a": 1, "b": 2}]) == {"a": "b"}
                return t(data)
                # directly return
            except (TypeError, ValueError):
                pass

        data = self._from_byte_like(self._attempt_from(data))

        if isinstance(data, str):
            try:
                return t(json.loads(data, strict=self.no_data_loss))  # noqa
            except json.decoder.JSONDecodeError:
                data = data.strip()
                if any(
                    data.startswith(v[0]) and data.endswith(v[1])
                    for v in self.STRUCTURE_BRACKET
                ):
                    import ast

                    res = self._attempt_from(ast.literal_eval(data))
                    if isinstance(res, dict):
                        # maybe set
                        return t(res)
                    raise
                if "=" in data:
                    if "&" in data:
                        # a=b&c=d   querystring index
                        from urllib.parse import parse_qs

                        qs = parse_qs(data)
                        return t({k: v[0] if len(v) == 1 else v for k, v in qs.items()})
                    spliter = ";" if ";" in data else ","
                    # cookie syntax or comma separate syntax
                    return t(
                        {
                            value.split("=")[0].strip(): ("=".join(value.split("=")[1:]).strip())
                            for value in data.split(spliter)
                        }
                    )
                raise

        from xml.etree.ElementTree import Element

        if isinstance(data, Element):
            return t(data.attrib)  # noqa

        return t(data)

    # bool is a subclass of int
    @registry.register(float)
    def to_float(self, data, t: Type[float] = float) -> float:
        if isinstance(data, float):
            return t(data)

        if self.no_explicit_cast:
            if not isinstance(data, (int, float, Decimal)):
                raise TypeError
        else:
            data = self._attempt_from_number(data)

        return t(data)

    @registry.register(int)
    def to_integer(self, data, t: Type[int] = int) -> int:
        if isinstance(data, int):
            # including bool:
            # True -> 1
            # False -> 0
            return t(data)

        if self.no_explicit_cast:
            if not isinstance(data, (int, float, Decimal)):
                raise TypeError
        else:
            data = self._attempt_from_number(data)
            if isinstance(data, str):
                if data.lower() in self.FALSE_VALUES:
                    return 0
                if data.lower() in self.TRUE_VALUES:
                    return 1

        try:
            data = Decimal(data)
        except decimal.InvalidOperation:
            raise TypeError(f'invalid number: {repr(data)}')
        # !!
        # FOR number > 1e+16, int(float()) will not get accurate result, use decimal instead

        if self.no_data_loss:
            if not data.is_normal():
                raise TypeError
            if data.as_tuple().exponent:
                raise TypeError

        return t(data)

    @registry.register(Decimal)
    def to_decimal(self, data, t: Type[Decimal] = Decimal) -> Decimal:
        if isinstance(data, Decimal):
            return t(data)  # noqa

        if self.no_explicit_cast:
            data = self._from_byte_like(data)
            if not isinstance(data, (int, float, str, Decimal)):
                raise TypeError
        else:
            data = self._attempt_from_number(data)

        return t(str(data).strip())  # noqa

    @registry.register(complex)
    def to_complex(self, data, t=complex) -> complex:
        if isinstance(data, t):
            return data

        if self.no_explicit_cast:
            data = self._from_byte_like(data)
            if not isinstance(data, (int, float, Decimal, str)):
                raise TypeError
        else:
            if isinstance(data, tuple) and len(data) == 2:
                return t(*data)

            data = self._attempt_from_number(data)

        return t(data)

    @registry.register(bool)  # after int
    def to_bool(self, data, t=bool) -> bool:
        if isinstance(data, bool):
            return t(data)
        if data == 1:
            return True
        elif data == 0:
            return False
        if self.no_explicit_cast:
            raise TypeError
        if isinstance(data, bytes):
            data = data.decode()
        rep = str(data).lower()
        if rep in self.FALSE_VALUES:
            return False
        elif rep in self.TRUE_VALUES:
            return True
        if self.no_data_loss:
            # bool can convert all the types
            # if the value is not bool (need to transform) in strict mode, throw an error directly
            raise TypeError
        return bool(data)

    @registry.register(
        date, allow_subclasses=False
    )  # datetime is a subclass of date
    def to_date(self, data, t: Type[date] = date) -> date:
        if isinstance(data, datetime):
            if self.no_data_loss:
                raise ValueError(f"Invalid date: {data}, must be date")
            return data.date()
        elif isinstance(data, t):
            return data

        dt = self.to_datetime(data, datetime)
        if self.no_data_loss:
            if dt.time() != time(0, 0):
                raise ValueError(f"Invalid date: {data}, got time part: {dt.time()}")
        return dt.date()

    @registry.register(datetime)
    def to_datetime(self, data, t: Type[datetime] = datetime) -> datetime:
        if isinstance(data, t):
            return data
        elif isinstance(data, date):
            return t(year=data.year, month=data.month, day=data.day)  # noqa

        data = self._attempt_from(data)
        try:
            num = self.to_float(data, float)
        except (TypeError, ValueError):
            pass
        else:
            while abs(num) > self.MS_WATERSHED:
                num /= 1000
            return t.utcfromtimestamp(num).replace(tzinfo=timezone.utc)

        data = self._from_byte_like(data)
        is_utc = "GMT" in data or 'UTC' in data or data.endswith("Z") and "T" in data
        data = data.replace('GMT', '').replace('UTC', '').replace('TZD', '').rstrip('Z').strip()

        for f in self.DATETIME_FORMATS:
            try:
                val = t.strptime(data, f)
                if is_utc:
                    val = val.replace(tzinfo=timezone.utc)
                return val
            except (TypeError, ValueError, re.error):
                continue

        if '+' in str(data):
            for f in self.DATETIME_FORMATS:
                try:
                    # val = t.strptime(data, f + ' %z')
                    val = t.strptime(data, f + (' %z' if ' +' in str(data) else '%z'))
                    if is_utc:
                        val = val.replace(tzinfo=timezone.utc)
                    return val
                except (TypeError, ValueError, re.error):
                    continue

        raise TypeError('invalid datetime')

    @registry.register(timedelta)
    def to_timedelta(self, data, t: Type[timedelta] = timedelta) -> timedelta:
        if isinstance(data, t):
            return data
        data = self._from_byte_like(self._attempt_from(data))
        try:
            num = self.to_float(data, float)
        except (TypeError, ValueError):
            pass
        else:
            if self.no_explicit_cast and isinstance(data, str):
                raise TypeError
            return t(seconds=num)
        if isinstance(data, str):
            for regex in self.DURATION_REGS:
                match = regex.match(data)
                if not match:
                    continue
                kw = match.groupdict()
                sign = -1 if kw.pop("sign", "+") == "-" else 1
                if kw.get("microseconds"):
                    kw["microseconds"] = kw["microseconds"].ljust(6, "0")

                if (
                    kw.get("seconds")
                    and kw.get("microseconds")
                    and kw["seconds"].startswith("-")
                ):
                    kw["microseconds"] = "-" + kw["microseconds"]
                kw_ = {k: float(v) for k, v in kw.items() if v is not None}
                return sign * t(**kw_)
            if self.no_explicit_cast:
                raise ValueError(f"Invalid timedelta: {data}")
            tm = time.fromisoformat(data)
            return t(
                hours=tm.hour,
                minutes=tm.minute,
                seconds=tm.second,
                microseconds=tm.microsecond,
            )
        raise TypeError

    @registry.register(time)
    def to_time(self, data, t: Type[time] = time) -> time:
        if isinstance(data, t):
            return data
        data = self._attempt_from(data)
        if not self.no_data_loss:
            if isinstance(data, datetime):
                return data.time()
            if isinstance(data, date):
                return t()
        data = self._from_byte_like(data)
        if isinstance(data, str):
            return t.fromisoformat(data)
        raise TypeError

    @registry.register(UUID)
    def to_uuid(self, data, t: Type[UUID] = UUID) -> UUID:
        if isinstance(data, t):
            return data

        if isinstance(data, str):
            return t(data)

        elif isinstance(data, (bytes, bytearray)):
            try:
                return t(data.decode())
            except ValueError:
                # 16 bytes in big-endian order as the bytes argument fail
                # the above check
                return t(bytes=data)

        if not self.no_explicit_cast:
            if not self.no_data_loss and isinstance(data, (float, Decimal)):
                data = int(data)

            if isinstance(data, int):
                return t(int=data)

        raise TypeError

    # enum should be the last of current types
    # because Enum subclass may inherit other base
    @registry.register(Enum)
    def to_enum(self, data, t: Type[Enum]) -> Enum:
        if isinstance(data, t):
            return data
        if self.no_explicit_cast:
            return t(data)  # noqa
        if not self.no_data_loss:
            if data in t.__members__:  # noqa
                return t.__members__[data]  # noqa
        member_type = getattr(t, "_member_type_", None)
        if member_type and member_type != object:
            if type(data) != member_type:
                data = self(data, member_type)
        return t(data)  # noqa

    @registry.register(io.BytesIO)
    def to_filelike(self, data, t):
        if isinstance(data, t):
            return data
        if isinstance(data, (bytes, bytearray, memoryview)):
            return t(data)
        if self.no_explicit_cast:
            raise TypeError
        return t(str(data).encode())

    @registry.register(type, allow_subclasses=False)
    def to_type(self, data, t=type) -> type:
        if not isinstance(data, type):
            raise TypeError('data is not a type')
        return data

    def handle_unresolved(self, data, t):
        if isinstance(t, type) and isinstance(data, t):
            # we just loosely match the isinstance for unresolved types
            return data
        if self.unresolved_types == "throw":
            raise TypeMismatchError(value=data, type=t)
        elif self.unresolved_types == "init":
            return t(data)
        return data

    def apply(self, data, t: Type[T], func=None) -> T:
        if not func:
            return self(data, t)
        if type(data) == t:
            # strict equal. not isinstance, like datetime is instance of date
            return data
        if isinstance(t, ForwardRef):
            if not t.__forward_evaluated__:
                raise TypeError(f"ForwardRef: {t} not evaluated")
            t = t.__forward_value__
        return func(self, data, t)

    def __call__(self, data, t: Type[T]) -> T:
        if isinstance(t, ForwardRef):
            if not t.__forward_evaluated__:
                raise TypeError(f"ForwardRef: {t} not evaluated")
            t = t.__forward_value__
        if type(data) == t:
            # strict equal. not isinstance, like datetime is instance of date
            return data
        transformer = self.resolver_transformer(t)
        if not transformer:
            return self.handle_unresolved(data, t)
        return transformer(self, data, t)


def type_transform(data, type: Type[T], options=None) -> T:
    from ..parser.options import Options

    context = (options or Options()).make_context()
    return context.transformer(data, type)



import warnings
from functools import partial
from typing import Callable, List, TypeVar, Union

from .parser.cls import ClassParser, init_dataclass
from .parser.field import ParserField
from .parser.options import Options, RuntimeContext
from .parser.rule import LogicalType
from .utils import exceptions as exc
from .utils.datastructures import unprovided

T = TypeVar("T")
OTHER = TypeVar("OTHER")


class LogicalMeta(type):
    __logical_type__ = LogicalType

    @property
    def __ref__(cls):
        try:
            name = cls.__qualname__
        except AttributeError:
            name = cls.__name__
        if "." in name:
            name = "".join(
                [
                    part.capitalize() if part.islower() else part
                    for part in name.split(".")
                ]
            )
        return f"{cls.__module__}.{name}"

    def __and__(cls: T, other: T) -> Union[T, OTHER]:
        if isinstance(other, LogicalType):
            return other.__rand__(cls)  # noqa
        return cls.__logical_type__.combine("&", cls, other)

    def __rand__(cls: T, other: OTHER) -> Union[OTHER, T]:
        return cls.__logical_type__.combine("&", other, cls)

    def __or__(cls: T, other: OTHER) -> Union[T, OTHER]:
        if getattr(other, "__origin__", None) == Union:
            return cls.__logical_type__.combine("|", cls, *other.__args__)
        if isinstance(other, LogicalType):
            return other.__ror__(cls)  # noqa
        return cls.__logical_type__.combine("|", cls, other)

    def __ror__(cls: T, other: OTHER) -> Union[OTHER, T]:
        if getattr(other, "__origin__", None) == Union:
            return cls.__logical_type__.combine("|", *other.__args__, cls)
        return cls.__logical_type__.combine("|", other, cls)

    def __xor__(cls: T, other: OTHER) -> Union[T, OTHER]:
        if isinstance(other, LogicalType):
            return other.__rxor__(cls)  # noqa
        return cls.__logical_type__.combine("^", cls, other)

    def __rxor__(cls: T, other: OTHER) -> Union[OTHER, T]:
        return cls.__logical_type__.combine("^", other, cls)

    def __invert__(cls):
        return cls.__logical_type__.combine("~", cls)


class DataClass(metaclass=LogicalMeta):
    __parser_cls__ = ClassParser
    __parser__: ClassParser
    __options__: Options

    def __init__(self, **kwargs):
        pass

    def __init_subclass__(cls, **kwargs):
        # options = getattr(cls, "__options__", None)
        cls.__parser__ = parser = cls.__parser_cls__.apply_for(cls)
        cls.__options__ = cls.__parser__.options

        parser.make_init(
            # init_super=False,
            # allow_runtime=True,
            # set_attributes=True,
            # coerce_property=False,
            post_init=cls.__post_init__,
        )
        parser.make_eq()
        parser.make_repr(ignore_str=False)
        parser.make_contains(output_only=True)
        parser.assign_properties(
            post_setattr=cls.__post_setattr__, post_delattr=cls.__post_delattr__
        )

    @property
    def __name__(self):
        return self.__parser__.name

    def __class_getitem__(cls, item):
        raise NotImplemented

    def __validate__(self):
        pass

    def __post_init__(self, values, context: RuntimeContext):  # noqa
        self.__validate__()

    def __post_setattr__(self, field: ParserField, value, context: RuntimeContext):
        pass

    def __post_delattr__(self, field: ParserField, context: RuntimeContext):
        pass

    # def __copy__(self):
    #     obj = self.__class__.__new__(self.__class__)
    #     obj.__dict__ = self.__dict__
    #     obj.__runtime_options__ = self.__runtime_options__
    #     return obj

    @classmethod
    def __from__(cls, data, options: Options = None):
        return init_dataclass(cls, data, options=options)

    def __export__(
        self,
        includes: Union[str, List[str]] = None,
        excludes: Union[str, List[str]] = None,
        as_attname: bool = False,
        no_output: Callable = None,  # control by value
        mode: str = None,
    ) -> dict:
        pass


class Schema(dict, metaclass=LogicalMeta):
    __parser_cls__ = ClassParser
    __parser__: ClassParser
    __options__: Options
    # __mode__: str = None

    def __init_subclass__(cls, **kwargs):
        # options = getattr(cls, "__options__", None)
        cls.__parser__ = parser = cls.__parser_cls__.apply_for(cls)
        cls.__options__ = cls.__parser__.options

        parser.make_init(
            # init_super=True,
            # allow_runtime=True,
            # set_attributes=False,   # n
            # coerce_property=True,
            post_init=cls.__post_init__,
        )
        parser.assign_properties(
            getter=cls.__field_getter__,
            setter=cls.__field_setter__,
            deleter=cls.__field_deleter__,
        )

        for key, field in parser.property_fields.items():
            # if field.property.fset:
            #   if field.always_no_output(cls.__options__.make_runtime(cls)) and not field.dependants:
            #       continue
            getter = partial(
                cls.__field_getter__, field=field, getter=field.property.fget
            )
            setter = (
                partial(cls.__field_setter__, field=field, setter=field.property.fset)
                if field.property.fset
                else None
            )
            deleter = (
                partial(cls.__field_deleter__, field=field, deleter=field.property.fdel)
                if field.property.fdel
                else None
            )

            for f in (getter, setter, deleter):
                if f:
                    f.__name__ = field.attname

            hooked_property = property(fget=getter, fset=setter, fdel=deleter)
            setattr(cls, field.attname, hooked_property)

    def __class_getitem__(cls, item):
        # if isinstance(item, Options):
        #     class _cls(cls):
        #         __options__ = item
        #     return _cls
        raise NotImplemented

    def __validate__(self):
        pass

    def __str__(self):
        items = []
        for key, val in self.items():
            field = self.__parser__.get_field(key)
            name = field.attname if field else key
            if field:
                repr_val = field.repr_value(val)
                if unprovided(repr_val):
                    # do not print
                    continue
            else:
                repr_val = repr(val)
            items.append(f"{name}={repr_val}")
        values = ", ".join(items)
        return f"{self.__name__}({values})"

    @property
    def __name__(self):
        return self.__parser__.name

    def __repr__(self):
        return self.__str__()

    @classmethod
    def __from__(cls, data, options=None):
        return init_dataclass(cls, data, options=options)

    # coerce_properties need to separate from set_attributes and execute by order
    # because the dependencies that the property need may not be set during one-time loop
    # (which is guarantee by the field orders, and consider not reliable)
    def __coerce_property__(self, field: ParserField, context: RuntimeContext):
        if field.always_no_output(context.options):
            return

        if field.dependencies and not field.dependencies.issubset(self):
            # maybe some of the dependencies is no_output=True, but still accessible through attribute
            # check if any of those dependencies is not in __dict__, and directly return if found one
            for dep in field.dependencies:
                dep_field = self.__parser__.get_field(dep)
                if not dep_field or dep_field.attname not in self.__dict__:
                    return

        try:
            attr = field.property.fget(self)  # get from the original getter
        except Exception as e:
            error_option = field.output_field.on_error if field.output_field else None
            msg = f"@property: {repr(field.attname)} calculate failed with error: {e}"
            if error_option == context.options.THROW:
                raise e.__class__(msg) from e
            else:
                warnings.warn(msg)
            return

        value = field.parse_output_value(  # parse @property result also
            attr, context=context
        )

        if unprovided(value):
            return

        if not field.is_no_output(value, options=context.options):
            super().__setitem__(field.name, value)
            # values[key] = value
            # do not apply cache here
            # when updating it will get nasty
            # _obj_self.__dict__[field.attname] = value
        else:
            if field.name in self:
                super().__delitem__(field.name)
            if field.attname in self.__dict__:
                self.__dict__.pop(field.attname)

        return value

    def __post_init__(self, values, context: RuntimeContext):
        super().__init__(values)
        self.__options__ = context.options  # set options
        for key, field in self.__parser__.property_fields.items():
            self.__coerce_property__(field, context=context)
        self.__validate__()
        context.raise_error()  # raise error if there is any

    def __contains__(self, item: str):
        field = self.__parser__.get_field(item)
        if not field:
            return super().__contains__(item)
        return super().__contains__(field.name)

    def __field_getter__(self, field: ParserField, getter: Callable = None):
        if field.name in self:
            # include output properties which is contained in the data
            return self[field.name]
        if field.attname in self.__dict__:
            # maybe a no_output field
            return self.__dict__[field.attname]

        if callable(getter):
            context = self.__parser__.make_context(force_error=True)
            value = field.parse_output_value(getter(self), context=context)
            if unprovided(value):
                raise AttributeError(
                    f"{self.__name__}: @property: {repr(field.attname)} failed to calculate"
                )
            return value

        deferred_default = field.get_default(
            options=self.__options__, defer=True
        )  # get deferred default
        if not unprovided(deferred_default):
            return deferred_default
        raise AttributeError(
            f"{self.__name__}: {repr(field.attname)} not provided in schema instance"
        )

    def __getitem__(self, item):
        # stay the same behaviour as the __contains__
        field = self.__parser__.get_field(item)
        if field:
            return super().__getitem__(field.name)
        return super().__getitem__(item)

    def __field_setter__(self, value, field: ParserField, setter: Callable = None):
        if self.__options__.immutable or field.immutable:
            raise exc.UpdateError(
                f"{self.__name__}: "
                f"Attempt to set immutable attribute: [{repr(field.attname)}]"
            )

        context = self.__parser__.make_context(force_error=True)
        value = field.parse_value(value, context=context)

        if field.property:
            if callable(setter):
                # @property.fset
                setter(self, value)

            # force calculate property
            self.__coerce_property__(field, context=context)
        else:
            if field.is_no_output(value, options=self.__options__):
                self.__dict__[field.attname] = value
                # no output
                if field.name in self:
                    super().__delitem__(field.name)
            else:
                super().__setitem__(field.name, value)

        if field.dependants:
            # need to update the dependant properties
            for dep in field.dependants:
                dep_field = self.__parser__.get_field(dep)
                if dep_field and dep_field.property:
                    self.__coerce_property__(dep_field, context=context)

    def __setitem__(self, alias: str, value):
        if self.__options__.immutable:
            raise exc.UpdateError(
                f"{self.__class__}: "
                f"Attempt to set item: [{repr(alias)}] in immutable schema"
            )

        field = self.__parser__.get_field(alias)
        if not field:
            if alias in self.__parser__.exclude_vars:
                raise exc.UpdateError(
                    f"{self.__class__}: Attempt to set excluded attribute: {repr(alias)}"
                )
            context = self.__parser__.make_context(force_error=True)
            addition = self.__parser__.parse_addition(alias, value, context=context)
            if unprovided(addition):
                # ignore addition
                return
            return super().__setitem__(alias, value)

        return self.__field_setter__(value, field=field)

    def __field_deleter__(self, field: ParserField, deleter: Callable = None):
        if self.__options__.immutable or field.immutable:
            raise exc.DeleteError(
                f"{self.__name__}: "
                f"Attempt to delete immutable attribute: [{repr(field.attname)}]"
            )

        if callable(deleter):
            deleter(self)

            if field.name in self:
                super().__delitem__(field.name)
        else:
            if field.is_required(self.__options__):
                raise exc.DeleteError(
                    f"{self.__name__}: Attempt to delete required schema key: {repr(field.attname)}"
                )
            if field.name not in self:
                raise exc.DeleteError(
                    f"{self.__name__}: Attempt to delete nonexistent attribute: {repr(field.attname)}"
                )
            super().__delitem__(field.name)

        if field.name in self.__dict__:
            self.__dict__.pop(field.attname)

    def __delitem__(self, key: str):
        if self.__options__.immutable:
            raise exc.DeleteError(
                f"{self.__class__}: "
                f"Attempt to delete item: [{repr(key)}] in immutable schema"
            )
        field = self.__parser__.get_field(key)
        if not field:
            return super().__delitem__(key)
        return self.__field_deleter__(field)

    def popitem(self):
        if self.__options__.immutable:
            raise exc.DeleteError(
                f"{self.__name__}: Attempt to popitem in immutable schema"
            )
        return super().popitem()

    def pop(self, key: str, default=unprovided):
        if self.__options__.immutable:
            raise exc.DeleteError(
                f"{self.__class__}: "
                f"Attempt to pop item: [{repr(key)}] in immutable schema"
            )
        field = self.__parser__.get_field(key)
        if not field:
            return super().pop(key)
        if field.immutable:
            raise exc.DeleteError(
                f"{self.__name__}: Attempt to pop immutable item: [{repr(key)}]"
            )
        if field.is_required(self.__options__):
            raise exc.DeleteError(
                f"{self.__name__}: Attempt to delete required schema key: {repr(key)}"
            )
        args = () if unprovided(default) else (default,)
        return super().pop(field.name, *args)

    def update(self, __m=None, **kwargs):
        if self.__options__.immutable:
            raise exc.UpdateError(
                f"{self.__name__}: Attempt to update in immutable schema"
            )
        data = dict(__m) if __m else kwargs
        for key, val in data.items():
            self.__setitem__(key, val)
        # TODO: reduce the dependant property calculation times if there are duplicate dependants

        # options = self.__options__.patch(
        #     ignored_options=['min_params'],
        #     ignore_required=True,
        #     no_default=True,
        # ).make_runtime()
        #
        # values = self.__parser__(
        #     # keep the original options settings including the addition
        #     dict(__m) if __m else kwargs, options=options
        # )
        #
        # self.__parser__.set_attributes(values, self, options=options)
        # self.__parser__.coerce_properties(values, self, options=options)
        #
        # return super().update(values)

    # def __copy__(self):
    #     return self.copy()

    def copy(self):
        obj = self.__class__.__new__(self.__class__)
        dict.update(obj, self)
        # since self.<data> is validated
        # we directly call dict.update to avoid calling the parsing methods again
        obj.__dict__ = self.__dict__
        return obj

    def clear(self):
        if self.__options__.immutable:
            raise exc.DeleteError(
                f"{self.__name__}: Attempt to clear in Options(immutable=True) schema"
            )
        for key, field in self.__parser__.fields.items():
            if field.immutable:
                raise exc.DeleteError(
                    f"{self.__name__}: Attempt to clear schema with immutable field: {repr(field.name)}"
                )
            if field.is_required(self.__options__):  # unless options is ignore_required
                raise exc.DeleteError(
                    f"{self.__name__}: Attempt to delete required schema key: {repr(key)}"
                )
        return super().clear()


DataClass.__init_subclass__()
Schema.__init_subclass__()



from setuptools import setup, find_packages
from utype import __version__

with open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

setup(
    name="utype",
    version=__version__,
    description="Declare & parse types / dataclasses / functions based on Python type annotations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="XuLin Zhou",
    author_email="zxl@utilmeta.com",
    keywords="utype type schema meta-type validation data-model type-transform parser json-schema",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet',
    ],
    install_requires=[
        'typing-extensions>=4.1.0',
    ],
    python_requires=">=3.7",
    license="Apache 2.0",
    url="https://utype.io",
    project_urls={
        "Project Home": "https://utype.io",
        "Documentation": "https://utype.io",
        "Source Code": "https://github.com/utilmeta/utype",
    },
    packages=find_packages(exclude=["tests.*", "tests", "docs.*"]),
)



import inspect
import warnings

from utype.parser.rule import Rule, LogicalType, SEQ_TYPES, MAP_TYPES
from utype.parser.field import ParserField
from utype.parser.cls import ClassParser
from utype.parser.func import FunctionParser
from utype.parser.base import Options
from decimal import Decimal
from datetime import datetime, date, time, timedelta
from uuid import UUID
from ipaddress import IPv4Address, IPv6Address
from typing import Optional, Type, Union, Dict
from ..utils.datastructures import unprovided
from ..utils.compat import JSON_TYPES
from enum import EnumMeta


class JsonSchemaGenerator:
    # pass in a defs dict to generate re-use '$defs'
    PRIMITIVES = ("null", "boolean", "object", "array", "integer", "number", "string")
    PRIMITIVE_MAP = {
        type(None): "null",
        bool: "boolean",
        MAP_TYPES: "object",
        SEQ_TYPES: "array",
        int: "integer",
        (float, Decimal): "number",
    }
    OPERATOR_NAMES = {
        "&": "allOf",
        "|": "anyOf",
        "^": "oneOf",
        "~": "not",
    }
    FORMAT_MAP = {
        (bytes, bytearray, memoryview): 'binary',
        float: 'float',
        IPv4Address: 'ipv4',
        IPv6Address: 'ipv6',
        datetime: 'date-time',
        date: 'date',
        time: 'time',
        timedelta: 'duration',
        UUID: 'uuid'
    }
    DEFAULT_CONSTRAINTS_MAP = {
        'enum': 'enum',
        'const': 'const',
    }
    TYPE_CONSTRAINTS_MAP = {
        ("integer", "number"): {
            'multiple_of': 'multipleOf',
            'le': 'maximum',
            'lt': 'exclusiveMaximum',
            'ge': 'minimum',
            'gt': 'exclusiveMinimum',
            'decimal_places': 'decimalPlaces',
            'max_digits': 'maxDigits',
            **DEFAULT_CONSTRAINTS_MAP,
        },
        ("array",): {
            'max_length': 'maxItems',
            'min_length': 'minItems',
            'unique_items': 'uniqueItems',
            'max_contains': 'maxContains',
            'min_contains': 'minContains',
            'contains': 'contains',
            **DEFAULT_CONSTRAINTS_MAP,
        },
        ("object",): {
            'max_length': 'maxProperties',
            'min_length': 'minProperties',
            **DEFAULT_CONSTRAINTS_MAP,
        },
        ("string",): {
            'regex': 'pattern',
            'max_length': 'maxLength',
            'min_length': 'minLength',
            **DEFAULT_CONSTRAINTS_MAP,
        },
        ("boolean", "null"): DEFAULT_CONSTRAINTS_MAP
    }

    FORMAT_PATTERNS = {
        'integer': r'[-]?\d+',
        'number': r'[-]?\d+(\.\d+)?',
        'date': r'\d{4}-\d{2}-\d{2}',
    }
    DEFAULT_PRIMITIVE = "string"
    DEFAULT_REF_PREFIX = "#/$defs/"

    def __init__(self, t,
                 defs: Dict[type, dict] = None,
                 names: Dict[str, type] = None,
                 ref_prefix: str = None,
                 mode: str = None,
                 output: bool = False
                 ):
        self.t = t
        self.defs = defs
        self.names = names
        if isinstance(defs, dict) and self.names is None:
            self.names = {}

        self.ref_prefix = ref_prefix or self.DEFAULT_REF_PREFIX
        self.mode = mode
        self.output = output
        self.options = Options(mode=mode)
        # can generate based on mode and input/output

    def generate_for_type(self, t: type):
        if t is None:
            return {}
        if not isinstance(t, type):
            return {}
        if issubclass(t, Rule):
            return self.generate_for_rule(t)
        elif isinstance(getattr(t, "__parser__", None), ClassParser):
            return self.generate_for_dataclass(t)
        elif isinstance(t, LogicalType) and t.combinator:
            return self.generate_for_logical(t)
        elif isinstance(t, EnumMeta):
            base = t.__base__
            enum_type = None
            enum_values = []
            enum_map = {}
            for key, val in t.__members__.items():
                enum_values.append(val.value)
                enum_map[key] = val.value
                enum_type = type(val.value)
            if not isinstance(base, EnumMeta):
                enum_type = base
            prim = self._get_primitive(enum_type)
            fmt = self._get_format(enum_type)
            data = {
                "type": prim,
                "enum": enum_values,
                "x-annotation": {
                    "enums": enum_map
                }
            }
            if fmt:
                data.update(format=fmt)
            return data

        # default common type
        prim = self._get_primitive(t)
        fmt = self._get_format(t)
        data = {"type": prim}
        if fmt:
            data.update(format=fmt)
        return data

    def generate_for_logical(self, t: LogicalType):
        operator_name = self.OPERATOR_NAMES.get(t.combinator)
        if not operator_name:
            return {}
        conditions = [self.generate_for_type(cond) for cond in t.args]
        if operator_name == 'not':
            return {operator_name: conditions[0]}
        return {operator_name: conditions}

    def _get_format(self, origin: type) -> Optional[str]:
        if not origin:
            return None
        format = getattr(origin, 'format', None)
        if format and isinstance(format, str):
            return format
        for types, f in self.FORMAT_MAP.items():
            if issubclass(origin, types):
                return f
        return None

    def _get_primitive(self, origin: type) -> str:
        if not origin:
            return self.DEFAULT_PRIMITIVE
        for types, pri in self.PRIMITIVE_MAP.items():
            if issubclass(origin, types):
                return pri
        return self.DEFAULT_PRIMITIVE

    def _get_args(self, r: Type[Rule]) -> dict:
        origin = r.__origin__
        args = r.__args__
        if not args:
            return {}
        if not origin:
            return {}
        args_res = [self.generate_for_type(arg) for arg in args]
        if issubclass(origin, tuple):
            if r.__ellipsis_args__:
                name = 'items'
                return {name: args_res[0]}
            else:
                name = 'prefixItems'
                return {name: args_res}
        elif issubclass(origin, SEQ_TYPES):
            name = 'items'
            return {name: args_res[0]}
        elif issubclass(origin, MAP_TYPES):
            name = 'patternProperties'
            key_arg: dict = args_res[0]
            val_arg: dict = args_res[1]
            pattern = dict(key_arg).get('pattern', None)
            if not pattern:
                fmt = key_arg.get('format') or key_arg.get('type')
                if fmt:
                    pattern = self.FORMAT_PATTERNS.get(fmt)
            pattern = pattern or '.*'
            return {name: {pattern: val_arg}}
        else:
            return {}

    def __call__(self) -> dict:
        if inspect.isfunction(self.t):
            return self.generate_for_function(self.t)
        return self.generate_for_type(self.t)

    def get_defs(self) -> Dict[str, dict]:
        defs = {}
        for t, values in self.defs.items():
            name = self.get_def_name(t)
            defs[name] = values
        return defs

    def generate_for_rule(self, t: Type[Rule]):
        name = t.__qualname__
        if isinstance(self.defs, dict):
            def_name = self.get_def_name(t)
            if def_name:
                return {"$ref": f"{self.ref_prefix}{def_name}"}

        # type
        origin = t.__origin__
        data = dict(self.generate_for_type(origin))
        primitive = getattr(t, 'primitive', None)
        if primitive in self.PRIMITIVES:
            data.update(type=primitive)
        else:
            primitive = data.get('type', self.DEFAULT_PRIMITIVE)

        # format
        fmt = getattr(t, 'format', None)
        if fmt and isinstance(fmt, str):
            data.update(format=fmt)
        else:
            fmt = self._get_format(origin)
            if fmt and isinstance(fmt, str):
                data.update(format=fmt)

        # constraints
        constrains_map = self.DEFAULT_CONSTRAINTS_MAP
        for types, mp in self.TYPE_CONSTRAINTS_MAP.items():
            if primitive in types:
                constrains_map = mp
                break
        for constraint, value, validator in t.__validators__:
            constraint_name = constrains_map.get(constraint, constraint)
            data[constraint_name] = value

        extra = getattr(t, 'extra', None)
        if extra and isinstance(extra, dict):
            data.update(extra)
        if t.__args__:
            data.update(self._get_args(t))
        if isinstance(self.defs, dict) and name != 'Rule':
            return {"$ref": f"{self.ref_prefix}{self.set_def(name, t, data)}"}
        return data

    def get_def_name(self, t: type):
        if t in self.defs:
            for k, v in self.names.items():
                if v == t:
                    return k
        return None

    def set_def(self, name: str, t: type, data: dict):
        n = 0
        while True:
            _name = name + (f'_{n}' if n else '')
            if _name in self.names:
                n += 1
                continue
            name = _name
            break
            # de-duplicate name
        self.defs[t] = data
        self.names[name] = t
        return name

    def generate_for_field(self, f: ParserField, options: Options = None) -> Optional[dict]:
        if self.output:
            if f.always_no_output(options or self.options):
                return None
        else:
            if f.always_no_input(options or self.options):
                return None

        t = f.output_type if self.output else f.type
        if not isinstance(t, type):
            if self.output:
                t = t or f.type
            else:
                t = t or f.output_type

        data = dict(self.generate_for_type(t))

        if f.field.title:
            data.update(title=f.field.title)
        if f.field.description:
            data.update(description=f.field.description)
        if f.field.deprecated:
            data.update(deprecated=f.field.deprecated)
        if f.field.mode:
            if f.field.mode == 'r':
                data.update(readOnly=True)
            elif f.field.mode == 'w':
                data.update(writeOnly=True)
        if not unprovided(f.field.example) and f.field.example is not None:
            example = f.field.example
            if type(f.field.example) not in JSON_TYPES:
                example = str(f.field.example)
            data.update(examples=[example])
        if f.aliases:
            aliases = list(f.aliases)
            if aliases:
                # sort to stay identical
                aliases.sort()
            data.update(aliases=aliases)
        annotations = f.schema_annotations
        if annotations:
            data.update({
                'x-annotation': annotations
            })
        return data

    # todo: de-duplicate generated schema class like UserSchema['a']
    def generate_for_dataclass(self, t):
        # name = t.__qualname__
        parser: ClassParser = getattr(t, '__parser__')
        cls_name = parser.name
        mode = parser.options.mode
        if mode:
            cls_name += '_' + mode
        if self.output and not parser.in_out_identical:
            cls_name += '_O'

        if isinstance(self.defs, dict):
            def_name = self.get_def_name(t)
            if def_name:
                return {"$ref": f"{self.ref_prefix}{def_name}"}

        data = {"type": "object"}
        required = []
        properties = {}

        for name, field in parser.fields.items():
            value = self.generate_for_field(field, options=parser.options)
            if value is None:
                continue
            properties[name] = value
            if field.is_required(parser.options or self.options):
                # will count options.ignore_required in
                required.append(name)
            elif self.output:
                if not field.no_default:
                    # if field has default, the value is required in the output data
                    required.append(name)

        data.update(properties=properties)
        if required:
            data.update(required=required)
        addition = parser.options.addition
        if addition is not None:
            if isinstance(addition, type):
                data.update(additionalProperties=self.generate_for_type(addition))
            else:
                data.update(additionalProperties=addition)

        annotations = parser.schema_annotations
        if annotations:
            data.update({
                'x-annotation': annotations
            })

        if isinstance(self.defs, dict):
            return {"$ref": f"{self.ref_prefix}{self.set_def(cls_name, t, data)}"}
        return data

    def generate_for_function(self, f):
        if not inspect.isfunction(f):
            raise TypeError(f'Invalid function: {f}')
        parser = getattr(f, '__parser__', None)
        if not isinstance(parser, FunctionParser):
            parser = FunctionParser.apply_for(f)
        data = {"type": "function"}
        pos_params = []
        required = []
        params = {}
        for name, field in parser.fields.items():
            value = self.generate_for_field(field, options=parser.options)
            if value is None:
                continue
            params[name] = value
            if field.positional_only:
                pos_params.append(name)
            if field.is_required(parser.options or self.options):
                required.append(name)
        data.update(parameters=params)
        if required:
            data.update(required=required)
        if pos_params:
            data.update(positionalOnly=pos_params)
        addition = parser.options.addition
        if addition is not None:
            if isinstance(addition, type):
                data.update(additionalParameters=self.generate_for_type(addition))
            else:
                data.update(additionalParameters=addition)
        return data

# REVERSE ACTION OF GENERATE:
# --- GENERATE Schema and types based on Json schema


class JsonSchemaParser:
    def __init__(self, json_schema: dict):
        pass



import inspect
import warnings
from collections.abc import Mapping
from datetime import date, datetime, time, timedelta, timezone
from decimal import Decimal
from ipaddress import IPv4Address, IPv6Address
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Union
from uuid import UUID

from ..utils import exceptions as exc
from ..utils.compat import Literal, get_args, is_final, is_annotated
from ..utils.datastructures import unprovided
from ..utils.functional import copy_value, get_name, multi
from .options import Options, RuntimeContext
from .rule import ConstraintMode, Lax, LogicalType, Rule, resolve_forward_type

represent = repr


class Field:
    parser_field_cls = None

    def __init__(
        self,
        *,
        alias: Union[str, Callable] = None,
        alias_from: Union[str, List[str], Callable, List[Callable]] = None,
        # can also be a generator
        case_insensitive: bool = None,
        # alias_for: str = None,     we may cancel alias for and replace it with alias
        required: Union[bool, str] = None,
        # required='rw'
        readonly: bool = None,
        # api def: cannot be part of the request body
        # this is the mark for upper-layer to apply
        # the schema-level readonly control is the "immutable" param
        writeonly: bool = None,
        mode: str = None,
        # api def: cannot be part of the response body
        # null: bool = False,
        # use null in type like Optional[int] / int | None
        default: Any = unprovided,
        default_factory: Callable = None,
        defer_default: bool = False,
        deprecated: Union[bool, str] = False,
        discriminator=None,  # discriminate the schema union by it's field
        no_input: Union[bool, str, Callable] = False,
        # can be a callable that takes the value of the field
        # give the bool of no_input / no_output
        # like no_output=lambda v: v is None  will ignore all None value when output
        no_output: Union[bool, str, Callable] = False,
        on_error: Literal["exclude", "preserve", "throw"] = None,  # follow the options
        # unprovided: Any = ...,
        immutable: bool = False,
        # secret: bool = None,    # None means depend on the Options.secret_names
        repr: Union[bool, str, Callable] = None,  # noqa
        dependencies: Union[list, str, property] = None,
        # (backup: internal, disallow, unacceptable)
        # unacceptable: we do not accept this field as an input,
        # but this field will be in the result if present (like default, or attribute set)
        # this is useful to auto_user_field or last_modified_field which we do not
        # want to expose the field to the user in the api docs
        # options: Options = None,  # for the annotated schema
        # property: _property = None,  # noqa
        # schema=None,
        # info: dict = None,
        # --- ANNOTATES ---
        title: str = None,
        description: str = None,
        example: Any = unprovided,
        # message: str = None,  # report this message if error occur
        # --- CONSTRAINTS ---
        const: Any = unprovided,
        enum: Iterable = None,
        gt=None,
        ge=None,
        lt=None,
        le=None,
        regex: str = None,
        length: Union[int, ConstraintMode] = None,
        max_length: Union[int, ConstraintMode] = None,
        min_length: int = None,
        # number
        max_digits: Union[int, ConstraintMode] = None,
        decimal_places: Union[int, ConstraintMode] = None,
        round: int = None,
        multiple_of: Union[int, ConstraintMode] = None,
        # array
        contains: type = None,
        max_contains: int = None,
        min_contains: int = None,
        unique_items: Union[bool, ConstraintMode] = None,
    ):
        if mode:
            if readonly or writeonly:
                raise exc.ConfigError(
                    f"Field: mode: ({represent(mode)}) cannot set with readonly or writeonly",
                    params={"mode": mode},
                )

        if readonly and writeonly:
            raise exc.ConfigError(
                f"Field: readonly and writeonly cannot be both specified",
                params={"readonly": readonly, "writeonly": writeonly},
            )

        if readonly:
            mode = "r"
        if writeonly:
            mode = "w"

        if deprecated:
            required = False

        if not unprovided(default):
            required = False
            if default_factory:
                raise exc.ConfigError(
                    f"Field: default: {represent(default)} and default factory cannot set both",
                    params={"default": default},
                )

        if default_factory:
            required = False
            if not callable(default_factory):
                raise exc.ConfigError(
                    f"Field: default_factory: {represent(default_factory)} must be a callable",
                    params={"default_factory": default_factory},
                )

        if defer_default:
            if unprovided(default) and not default_factory:
                raise exc.ConfigError(
                    f"Field: specify defer_default=True without any default",
                    params={
                        "defer_default": defer_default,
                        "default": default,
                        "default_factory": default_factory,
                    },
                )

        if required is None:
            required = True

        if required:
            if on_error == "exclude":
                raise exc.ConfigError(
                    f'Field: required field does not support on_error="exclude"',
                    params={"required": required, "on_error": on_error},
                )

            if isinstance(required, str):
                if mode:
                    if not set(required).issubset(set(mode)):
                        raise exc.ConfigError(
                            f"Field required: {represent(required)} is not in mode: {represent(mode)}",
                            params={"required": required, "mode": mode},
                        )

        if isinstance(no_input, str):
            if mode:
                if not set(no_input).issubset(set(mode)):
                    raise exc.ConfigError(
                        f"Field no_input: {represent(no_input)} is not in mode: {represent(mode)}",
                        params={"no_input": no_input, "mode": mode},
                    )

        if isinstance(no_output, str):
            if mode:
                if not set(no_output).issubset(set(mode)):
                    raise exc.ConfigError(
                        f"Field no_output: {represent(no_output)} is not in mode: {represent(mode)}",
                        params={"no_output": no_output, "mode": mode},
                    )

        if round:
            if decimal_places and decimal_places not in (round, Lax(round)):
                raise exc.ConfigError(
                    f"Field round: {round} is a shortcut for decimal_places=Lax({round}), "
                    f"but you specified a different decimal_places: {represent(decimal_places)}",
                    params={"round": round, "decimal_places": decimal_places},
                )

            decimal_places = Lax(round)

        self.alias = alias if isinstance(alias, str) else None
        self.alias_generator = alias if callable(alias) else None
        self.alias_from = alias_from
        # self.alias_from =
        self.case_insensitive = case_insensitive
        self.deprecated = bool(deprecated)
        self.deprecated_to = deprecated if isinstance(deprecated, str) else None

        self.no_input = no_input
        self.no_output = no_output
        self.immutable = immutable
        self.required = required

        self.default = default
        self.default_factory = default_factory
        self.defer_default = defer_default

        # self.unprovided = unprovided

        deps = []
        if dependencies:
            if not multi(dependencies):
                dependencies = [dependencies]

            for dep in dependencies:
                dep = get_name(dep)
                if not isinstance(dep, str):
                    raise exc.ConfigError(
                        f"Invalid dependency: {represent(dep)}, must be str or property"
                    )
                deps.append(dep)

        if discriminator:
            discriminator = get_name(discriminator)
            if not isinstance(discriminator, str):
                raise exc.ConfigError(
                    f"Field.discriminator must be a str, got {represent(discriminator)}",
                    params={"discriminator": discriminator},
                )

        self.dependencies = deps
        self.discriminator = discriminator
        self.on_error = on_error
        self.mode = mode

        self.title = title
        self.description = description
        self.example = example
        # self.message = message
        # self.secret = secret  # will display "******" instead of real value in repr

        if isinstance(repr, bool):
            if repr:
                repr = represent  # noqa
            else:
                repr = lambda v: unprovided  # noqa
        elif isinstance(repr, str):

            def repr(v, _=repr):  # noqa
                return _

        self.repr: Optional[Callable] = repr

        constraints = {
            k: v
            for k, v in dict(
                enum=enum,
                gt=gt,
                ge=ge,
                lt=lt,
                le=le,
                min_length=min_length,
                max_length=max_length,
                length=length,
                regex=regex,
                max_digits=max_digits,
                decimal_places=decimal_places,
                multiple_of=multiple_of,
                contains=contains,
                max_contains=max_contains,
                min_contains=min_contains,
                unique_items=unique_items,
            ).items()
            if v is not None
        }
        if not unprovided(const):
            constraints.update(const=const)

        self.constraints = constraints

    def get_alias(self, attname: str, generator=None):
        alias = attname
        if self.alias:
            alias = self.alias
        else:
            generator = self.alias_generator or generator
            if generator:
                _alias = generator(attname)
                if isinstance(_alias, str) and _alias:
                    alias = _alias
        return alias

    def get_alias_from(self, attname: str, generator=None) -> Set[str]:
        aliases = {attname}
        alias_from = []
        if self.alias_from:
            if not multi(self.alias_from):
                alias_from = [self.alias_from]
            else:
                alias_from = self.alias_from
            # field alias_from will override generator
        elif generator:
            if multi(generator):
                alias_from.extend(generator)
            else:
                alias_from.append(generator)

        for alias in alias_from:
            if callable(alias):
                alias = alias(attname)
            if multi(alias):
                aliases.update([a for a in alias if isinstance(a, str) and a])
            elif isinstance(alias, str) and alias:
                aliases.add(alias)

        return aliases

    # def to_spec(self):
    #     # convert to schema specification
    #     # like https://json-schema.org/
    #     pass

    def __call__(self, fn_or_cls, *args, **kwargs):
        setattr(fn_or_cls, "__field__", self)
        return fn_or_cls

    @property
    def schema_annotations(self):
        return {}

    @property
    def default_type(self):
        return None


class Param(Field):
    def __init__(
        self,
        default=unprovided,
        *,
        default_factory: Callable = None,
        alias: Union[str, Callable] = None,
        alias_from: Union[str, List[str], Callable, List[Callable]] = None,
        case_insensitive: bool = None,
        mode: str = None,
        deprecated: Union[bool, str] = False,
        discriminator=None,  # discriminate the schema union by it's field
        no_input: Union[bool, str, Callable] = False,
        on_error: Literal["exclude", "preserve", "throw"] = None,  # follow the options
        dependencies: Union[list, str, property] = None,
        # --- ANNOTATES ---
        title: str = None,
        description: str = None,
        example=unprovided,
        # --- CONSTRAINTS ---
        const=unprovided,
        enum: Iterable = None,
        gt=None,
        ge=None,
        lt=None,
        le=None,
        regex: str = None,
        length: Union[int, ConstraintMode] = None,
        max_length: Union[int, ConstraintMode] = None,
        min_length: int = None,
        # number
        max_digits: Union[int, ConstraintMode] = None,
        decimal_places: Union[int, ConstraintMode] = None,
        round: int = None,
        multiple_of: Union[int, ConstraintMode] = None,
        # array
        contains: type = None,
        max_contains: int = None,
        min_contains: int = None,
        unique_items: Union[bool, ConstraintMode] = None,
    ):
        kwargs = dict(locals())
        kwargs.pop("self")
        kwargs.pop("__class__")
        super().__init__(**kwargs)


class ParserField:
    TYPE_PRIMITIVE = {
        str: "string",
        int: "number",
        float: "number",
        bool: "boolean",
        list: "array",
        tuple: "array",
        set: "array",
        frozenset: "array",
        dict: "object",
        bytes: "string",
        Decimal: "number",
        date: "string",
        time: "string",
        datetime: "string",
        UUID: "string",
        Mapping: "object",
    }

    TYPE_FORMAT = {
        int: "integer",
        float: "float",
        tuple: "tuple",
        set: "set",
        frozenset: "set",
        bytes: "binary",
        Decimal: "decimal",
        date: "date",
        time: "time",
        datetime: "date-time",
        UUID: "uuid",
        timedelta: "duration",
        timezone: "timezone",
        IPv4Address: "ipv4",
        IPv6Address: "ipv6",
    }

    SECRET_EXEMPT_TYPES = (bool,)
    SECRET_REPR = "*" * 6

    field_cls = Field
    rule_cls = Rule
    # transformer_cls = TypeTransformer

    def __init__(
        self,
        name: str,
        # the actual schema field name
        # alias / attname
        input_type: type,
        # all the transformers and validators are infer from type
        field: Field,
        attname: str = None,
        aliases: Set[str] = None,
        field_property: property = None,
        output_type: type = None,
        output_field: Field = None,
        dependencies: Set[str] = None,
        final: bool = False,
        positional_only: bool = False,
        default=unprovided,     # passed when using Annotated[]
        **kwargs
    ):

        self._kwargs = kwargs

        self.attname = attname
        self.type = input_type
        self.output_type = output_type
        self.field = field
        self.output_field = output_field
        self.property = field_property
        self.final = final
        self.name = name
        self.aliases = set(aliases or []).difference({self.name})
        self.all_aliases = self.aliases.union({self.name})

        # self.input_transformer = self.transformer_cls.resolver_transformer(input_type)
        self.dependencies = dependencies
        self.attr_dependencies = set()
        self.dependants = set()

        # self.secret = field.secret
        # self.secret_repr = field.secret if isinstance(field.secret, str) else self.SECRET_REPR
        self.input_transformer = None
        self.output_transformer = None
        self.const = unprovided
        self.discriminator_map = {}
        self.positional_only = positional_only

        # ----------
        # below is the copy of the original field
        # which can be tweaked by subclasses
        self.deprecated_to = None
        self.repr_func = self.field.repr
        self.mode = self.field.mode
        self.no_input = self.field.no_input
        self.no_output = self.output_field.no_output if self.output_field else self.field.no_output
        self.default_factory = self.field.default_factory
        self.default = default if not unprovided(default) else self.field.default
        self.required = False if not unprovided(default) else self.field.required
        self.defer_default = self.field.defer_default
        self.on_error = self.field.on_error
        self.case_insensitive = self.field.case_insensitive
    # ----------------
    # below are static field properties

    @property
    def discriminator(self):
        return self.field.discriminator

    @property
    def alias(self):
        return self.field.alias

    @property
    def immutable(self):
        if self.final:
            return True
        return self.field.immutable

    @property
    def no_default(self):
        return unprovided(self.default) and not self.default_factory

    @property
    def input_origins(self) -> List[type]:
        if isinstance(self.type, LogicalType):
            return self.type.resolve_origins()
        elif isinstance(self.type, type):
            return [self.type]
        return []  # not set or ForwardRef

    @property
    def output_origins(self) -> List[type]:
        if isinstance(self.output_type, LogicalType):
            return self.output_type.resolve_origins()
        elif isinstance(self.output_type, type):
            return [self.output_type]
        return []  # not set or ForwardRef

    def repr_value(self, value):
        if self.repr_func:
            rp = self.repr_func(value)
            if unprovided(rp):
                return rp
            return str(rp)
        return represent(value)

    def setup(self, options: Options):
        if self.is_case_insensitive(options):
            # do not lower name
            # self.name = self.name.lower()
            self.aliases = {a.lower() for a in self.aliases}
            self.all_aliases = {a.lower() for a in self.all_aliases}

        if self.repr_func is None:
            if options.secret_names:
                is_secret = False
                for name in options.secret_names:
                    if name in self.all_aliases:
                        is_secret = True
                        break
                if is_secret:
                    # some types like bool is not some type to hide even if name in the secret names
                    origins = self.input_origins
                    if any([ori in self.SECRET_EXEMPT_TYPES for ori in origins]):
                        is_secret = False

                if is_secret:
                    self.repr_func = lambda v: self.SECRET_REPR

        if isinstance(self.type, type):
            if issubclass(self.type, Rule):
                self.const = getattr(self.type, "const", unprovided)
            else:
                trans = Rule.transformer_cls.resolver_transformer(self.type)
                if not trans:
                    if options.unresolved_types == options.THROW:
                        warnings.warn(
                            f"Field(name={repr(self.name)}) got unresolved type: {self.type}, "
                            f'and Options.unresolved_types == "throw", which will raise error'
                            f" in the runtime if the input value type does not match"
                        )
                else:
                    self.input_transformer = trans

        if isinstance(self.output_type, type):
            if issubclass(self.output_type, Rule):
                pass
            else:
                trans = Rule.transformer_cls.resolver_transformer(self.output_type)
                if not trans:
                    if options.unresolved_types == options.THROW:
                        warnings.warn(
                            f"Field(name={repr(self.name)}) got unresolved output type: {self.output_type}, "
                            f'and Options.unresolved_types == "throw", which will raise error'
                            f" in the runtime if the input value type does not match"
                        )
                else:
                    self.output_transformer = trans

        if self.discriminator:
            discriminator_map = {}
            comb = None
            if isinstance(self.type, LogicalType):
                comb = self.type.resolve_combined_origin()

            if not comb:
                raise TypeError(
                    f"Field: {repr(self.attname)} specify a discriminator: "
                    f"{repr(self.discriminator)}, but got a common type: {self.type} "
                    f"which does not support discriminator"
                )

            if comb.combinator == "|" or comb.combinator == "^":
                from .cls import ClassParser

                for arg in comb.args:
                    if arg is type(None):
                        # we accept None
                        continue

                    cls_parser = ClassParser.resolve_parser(arg)
                    if not cls_parser:
                        raise exc.ConfigError(
                            f"Field: {repr(self.attname)} specify a discriminator: "
                            f"{repr(self.discriminator)}, but got a type: {arg} "
                            f"that not support, must be a data class "
                            f"(like subclass of DataClass / Schema or "
                            f"use @dataclass to decorate a class)",
                            field=self.name,
                        )

                    field = cls_parser.get_field(self.discriminator)
                    if not isinstance(field, ParserField):
                        raise exc.ConfigError(
                            f"Field: {repr(self.attname)} specify a discriminator: "
                            f"{repr(self.discriminator)}, but is was not find in type: "
                            f"{arg}, you should define {self.discriminator}: "
                            f'Literal["some-value"] in that schema',
                            field=self.name,
                        )

                    const = field.const
                    if not isinstance(const, (int, str, bool)):
                        raise exc.ConfigError(
                            f"Field: {repr(self.attname)} specify a discriminator: "
                            f"{repr(self.discriminator)}, but in type {arg}, there is no"
                            f" common type const ({repr(const)}) set for this field, you should "
                            f"define {self.discriminator}: "
                            f'Literal["some-value"] in that schema',
                            field=self.name,
                        )

                    if const in discriminator_map:
                        raise exc.ConfigError(
                            f"Field: {repr(self.attname)} with discriminator: "
                            f"{repr(self.discriminator)}, got a duplicate value:"
                            f" {repr(const)} for {arg} and {discriminator_map[const]}",
                            field=self.name,
                        )

                    discriminator_map[const] = arg
                self.discriminator_map = discriminator_map
            else:
                raise TypeError(
                    f"Field: {repr(self.attname)} specify a discriminator: "
                    f"{repr(self.discriminator)}, but got a logical type: {self.type} "
                    f"with combinator: {repr(comb.combinator)} which does not support discriminator, "
                    f'only "^"(OneOf) or "|"(AnyOf) support'
                )

    def add_dependant(self, name: str):
        if name in self.dependants:
            return
        self.dependants.add(name)

    def apply_fields(
        self,
        fields: Dict[str, "ParserField"],
        # excluded_vars: Set[str],
        alias_map: dict,
        # attr_alias_map: dict,
    ):
        """
        take the field
        """
        if self.aliases:
            inter = self.aliases.intersection(fields)
            if inter:
                raise exc.ConfigError(
                    f"Field(name={repr(self.name)}) aliases: {inter} conflict with fields"
                )

        if self.dependencies:
            dependencies = []
            attr_dependencies = []
            for dep in self.dependencies:
                if dep in alias_map:
                    dep = alias_map[dep]
                if dep not in fields:
                    # continue
                    # if dependencies is generated from unbound, it is considered inaccurate
                    if not self.property:
                        raise exc.ConfigError(
                            f"Field(name={repr(self.name)}) dependency: {repr(dep)} not exists"
                        )
                    continue

                field = fields[dep]
                if self.property:
                    # if no getter function
                    # dependant will not affect
                    field.add_dependant(self.name)
                if dep not in dependencies:
                    dependencies.append(dep)
                if field.attname not in attr_dependencies:
                    attr_dependencies.append(field.attname)
            self.dependencies = set(dependencies)
            self.attr_dependencies = set(attr_dependencies)

        if self.field.deprecated_to:
            to = self.field.deprecated_to
            if to in alias_map:
                to = alias_map[to]
            if to not in fields:
                raise exc.ConfigError(
                    f"Field(name={repr(self.name)}) is deprecated,"
                    f" but prefer field : {repr(to)} not exists"
                )
            self.deprecated_to = to

    def resolve_forward_refs(self):
        if self.type:
            self.type, r = resolve_forward_type(self.type)
        if self.output_type:
            self.output_type, r = resolve_forward_type(self.output_type)

    @property
    def always_provided(self):
        # required is mode str or callable does not means
        return self.required is True or not self.no_default

    def is_case_insensitive(self, options: Options) -> bool:
        if self.case_insensitive is not None:
            return bool(self.case_insensitive)
        return bool(options.case_insensitive)

    # def get_unprovided(self, options: Options):
    #     # options = options or self.options
    #     if options and options.unprovided_attribute is not ...:
    #         value = options.unprovided_attribute
    #     elif self.field.unprovided is not ...:
    #         value = self.field.unprovided
    #     else:
    #         return ...
    #     if callable(value):
    #         return value()
    #     return copy_value(value)

    def get_default(self, options: Options, defer: Optional[bool] = False):
        # options = options or self.options
        if options.no_default:
            return unprovided

        if isinstance(defer, bool):
            if not defer:
                if self.defer_default or options.defer_default:
                    return unprovided
            else:
                if not self.defer_default and not options.defer_default:
                    return unprovided

        if not unprovided(options.force_default):
            default = options.force_default
        elif not unprovided(self.default):
            default = self.default
        elif self.default_factory:
            try:
                default = self.default_factory()
            except Exception as e:
                # we should directly raise the error since it is a "ServerError" instead of a parse error
                # we just want to add some info here to help debug
                raise e.__class__(
                    f"Field(name={repr(self.name)}) generate default failed with error: {e}"
                ) from e
        else:
            return unprovided
        return copy_value(default)

    def get_on_error(self, options: Options):
        if self.on_error:
            return self.on_error
        return options.invalid_values

    def is_required(self, options: Options):
        if options.ignore_required or not self.required:
            return False
        if self.always_no_input(options):
            return False
        if self.required is True:
            return True
        if not options.mode:
            return False
        return options.mode in self.required

    def is_no_input(self, value, options: Options):
        if self.final:
            if not self.no_default:
                return True

        no_input = (
            self.no_input(value)
            if callable(self.no_input)
            else self.no_input
        )

        if not options.mode:
            # no mode
            return no_input if isinstance(no_input, bool) else False

        if isinstance(no_input, (str, list, set, tuple)):
            return options.mode in no_input

        if no_input is True:
            return True

        if self.mode:
            return options.mode not in self.mode

        return bool(no_input)

    def always_no_input(self, options: Options):
        # calculate before get the value
        if self.final:
            if not self.no_default:
                return True
        if self.no_input is True:
            return True
        if not options.mode:
            return False
        if callable(self.no_input):
            return False
        if isinstance(self.no_input, (str, list, set, tuple)):
            if options.mode in self.no_input:
                return True
        if self.mode:
            return options.mode not in self.mode
        return False

    def always_no_output(self, options: Options):
        # calculate before get the value
        if self.no_output is True:
            return True
        if not options.mode:
            return False
        if callable(self.no_output):
            return False
        if isinstance(self.no_output, (str, list, set, tuple)):
            if options.mode in self.no_output:
                return True
        if self.mode:
            return options.mode not in self.mode
        return False

    def is_no_output(self, value, options: Options):
        # field = self.output_field or self.field
        # prefer the config in output field rather than input field
        no_output = (
            self.no_output(value)
            if callable(self.no_output)
            else self.no_output
        )

        if not options.mode:
            # no mode
            return no_output if isinstance(no_output, bool) else False

        if isinstance(no_output, (str, list, set, tuple)):
            return options.mode in no_output

        if no_output is True:
            return True

        if self.mode:
            return options.mode not in self.mode

        return bool(no_output)

    def check_function(self, func):
        if not self.always_provided:
            raise exc.ConfigError(
                f"{func}: Field(name={repr(self.name)}) in function with required=False "
                f"must specify a default or default_factory",
                obj=func,
                field=self.name,
                params={"required": self.required, "default": self.default},
            )

        if self.defer_default:
            # use defer_default=True will cause no default provided
            raise exc.ConfigError(
                f"{func}: Field(name={repr(self.name)}).defer_default has no meanings in function params,"
                f" please consider move it",
                obj=func,
                field=self.name,
                params={"defer_default": self.defer_default},
            )

        if self.no_default:
            if self.no_input:
                raise exc.ConfigError(
                    f"{func}: Field(name={repr(self.name)}) with no_input ({self.no_input}) "
                    f"must specify a default/default_factory value",
                    obj=func,
                    field=self.name,
                    params={
                        "no_input": self.no_input,
                        "default": self.default,
                    },
                )

            if self.mode:
                raise exc.ConfigError(
                    f"{func}: Field(name={repr(self.name)}) with mode ({self.mode}) "
                    f"must specify a default/default_factory value",
                    obj=func,
                    field=self.name,
                    params={"mode": self.mode, "default": self.default},
                )

        if self.positional_only:
            if self.field.alias:
                warnings.warn(
                    f"{func}: Field(name={repr(self.name)}).alias ({repr(self.field.alias)}) "
                    f"has no meanings in positional only params,"
                    f" please consider move it"
                )
            if self.field.alias_from:
                warnings.warn(
                    f"{func}: Field(name={repr(self.name)}).alias_from ({repr(self.field.alias_from)}) "
                    f"has no meanings in positional only params,"
                    f" please consider move it"
                )
            if self.case_insensitive:
                warnings.warn(
                    f"{func}: Field(name={repr(self.name)}).case_insensitive "
                    f"has no meanings in positional only params,"
                    f" please consider move it"
                )

        if self.no_output:
            warnings.warn(
                f"{func}: Field(name={repr(self.name)}).no_output has no meanings in function params,"
                f" please consider move it"
            )

        if self.immutable:
            warnings.warn(
                f"{func}: Field(name={repr(self.name)}).immutable has no meanings in function params, "
                f"please consider move it"
            )

        if self.field.repr:
            warnings.warn(
                f"{func}: Field(name={repr(self.name)}).repr has no meanings in function params,"
                f" please consider move it"
            )

    # 1. deal with parse: use context
    def parse_output_value(self, value, context: RuntimeContext):
        type = self.output_type
        if not type:
            return value
        trans = context.transformer
        try:
            return trans(value, type)  # noqa
        except Exception as e:
            error = exc.ParseError(
                item=self.name,
                type=self.output_type,
                value=value,
                field=self,
                origin_exc=e,
            )
            # todo: apply and distinct input field / output field
            error_option = (
                self.output_field.on_error if self.output_field else None
            ) or context.options.invalid_values
            if error_option == context.options.EXCLUDE:
                context.collect_waring(error.formatted_message)
            elif error_option == context.options.PRESERVE:
                context.collect_waring(error.formatted_message)
                return value
            else:
                context.handle_error(error)
            return unprovided

    def parse_value(self, value, context: RuntimeContext):
        if self.field.deprecated:
            to = (
                f", use {repr(self.deprecated_to)} instead"
                if self.deprecated_to
                else ""
            )
            context.collect_waring(
                f"{repr(self.name)} is deprecated{to}", category=DeprecationWarning
            )

        type = self.type
        # trans = context.transformer

        if self.discriminator_map and value is not None:
            if not isinstance(value, Mapping):
                try:
                    value = context.transformer.to_dict(value)
                except Exception as e:
                    context.handle_error(
                        exc.ParseError(
                            item=self.name,
                            type=dict,
                            value=value,
                            field=self,
                            origin_exc=e,
                        )
                    )
                    return unprovided

            discriminator = value.get(self.discriminator)
            if discriminator in self.discriminator_map:
                type = self.discriminator_map[discriminator]
                # directly assign type instead parse it in a Logical context
            else:
                context.handle_error(
                    exc.DiscriminatorMismatchError(
                        discriminator=self.discriminator,
                        discriminator_value=discriminator,
                        field=self,
                        value=value,
                        item=self.name,
                        type=self.type,
                    )
                )
                return unprovided

        if not type:
            # type is None, not type(None), means the exact same as Any / Rule
            return value

        with context.enter(self.name) as new_context:
            try:
                return new_context.transformer(value, type)  # noqa
            except Exception as e:
                error = exc.ParseError(
                    item=self.name,
                    type=self.type,
                    value=value,
                    field=self,
                    origin_exc=e,
                )
                error_option = self.get_on_error(context.options)
                if error_option == context.options.EXCLUDE:
                    if self.is_required(context.options):
                        # required field cannot be excluded
                        context.handle_error(error)
                    else:
                        context.collect_waring(error.formatted_message)
                    # return default if provided
                    # return unprovided if no default is set
                    return self.get_default(options=context.options, defer=False)
                elif error_option == context.options.PRESERVE:
                    context.collect_waring(error.formatted_message)
                    return value
                else:
                    context.handle_error(error)
                return unprovided

    @classmethod
    def process_annotate_meta(cls, m, **kwargs):
        if isinstance(m, Field):
            return m
        elif inspect.isclass(m) and issubclass(m, Field):
            return m()
        return None

    @classmethod
    def get_field(cls, annotation: Any, default, **kwargs):
        if unprovided(default):
            if isinstance(annotation, type):
                # if field use @field to decorate a type
                # and using this type like <var: t> without a default
                # then the __field__ prop will be detected and used as the field
                return getattr(annotation, '__field__', None)
        else:
            return default

    @property
    def schema_annotations(self):
        # this is meant to be extended and override
        # if the result is not None, it will become the x-annotation of the JSON schema output
        return self.field.schema_annotations

    @classmethod
    def generate(
        cls,
        attname: str,
        annotation: Any,
        default: Any,
        options: Options,
        positional_only: bool = False,
        global_vars=None,
        forward_refs=None,
        force_clear_refs=False,
        **kwargs
    ):
        prop = None
        output_type = None
        dependencies = None
        field = cls.get_field(annotation, default, **kwargs)
        output_field = None
        field_kwargs = dict()

        if isinstance(default, property):
            prop = default
            default = unprovided
            if prop.fset:
                _, (k, param) = inspect.signature(prop.fset).parameters.items()
                param: inspect.Parameter
                if param.annotation != param.empty:
                    annotation = param.annotation

                field = getattr(prop.fset, "__field__", None)

                if param.default != param.empty:
                    # @property
                    # @Field(...)
                    # def prop(self):
                    #     pass
                    #
                    # @prop.setter
                    # def prop(self, value: str = Field(...)):
                    #     pass

                    # default = param.default
                    if isinstance(param.default, Field):
                        field = param.default

                        # some invalid configures
                        if field.alias:
                            raise exc.ConfigError(
                                field=attname, params={"alias": field.alias}
                            )
                        if field.no_output:
                            raise exc.ConfigError(
                                field=attname, params={"no_output": field.no_output}
                            )
                        if field.dependencies:
                            raise exc.ConfigError(
                                field=attname,
                                params={"dependencies": field.dependencies},
                            )

                    else:
                        default = param.default
                        # raise ValueError(
                        #     f"property: {repr(attname)} defines Field i"
                        #     f"n setter param default value, which is not appropriate, "
                        #     f"you should use @{default} over the @property"
                        # )
            else:
                field_kwargs.update(
                    no_input=True,
                    required=False
                )

            if prop.fget:
                return_annotation = getattr(prop.fget, "__annotations__", {}).get(
                    "return"
                )

                # dependencies = inspect.getclosurevars(prop.fget).unbound
                # !!
                # UNBOUND MAY GET THE WRONG DEPENDENCY AND THUS THE WRONG RESULT (without explicit error)
                # [SUCH AS self.a.b will make "b" an unbound, which is not the desired dependency]
                # SO WE SHOULD LET DEVELOPER SPECIFY DEPENDENCIES OR [LET IT FAIL]

                # use the unbound properties as default dependencies of property
                # you can use @Field(dependencies=[...]) to specify yourself

                output_field = getattr(prop.fget, "__field__", None)
                if isinstance(output_field, Field):

                    if output_field.dependencies:
                        dependencies = output_field.dependencies

                    # some invalid configures
                    if output_field.no_input:
                        raise exc.ConfigError(
                            field=attname, params={"no_input": output_field.no_input}
                        )
                    if output_field.alias_from:
                        raise exc.ConfigError(
                            field=attname, params={"alias_from": output_field.alias_from}
                        )
                    if not unprovided(output_field.default) or output_field.default_factory:
                        raise exc.ConfigError(
                            field=attname, params={"default": output_field.default}
                        )

                else:
                    output_field = None

                output_type = cls.rule_cls.parse_annotation(
                    annotation=return_annotation,
                    global_vars=global_vars,
                    forward_refs=forward_refs,
                    forward_key=attname,
                    constraints=output_field.constraints if output_field else None,
                    force_clear_refs=force_clear_refs
                )

            else:
                field_kwargs.update(no_output=True)

        final = is_final(annotation)
        if final:
            # turn to Any by default, not bare Final, which will not be recognized as a valid annotation
            # _origin = get_origin(annotation)
            # if _origin == Final:
            # this type should take care in this level
            # because it does not affect validation / transformation
            # and rather a field behaviour
            args = get_args(annotation)
            if args:
                annotation = args[0]
            else:
                annotation = Any

        _default = unprovided
        if is_annotated(annotation):
            # Annotated[origin, field]
            field = None
            _default = default
            # Annotated[t, f] = d
            # makes d the absolute default, even for Field instance
            for m in getattr(annotation, '__metadata__', []):
                field = cls.process_annotate_meta(m, **kwargs)
                if field:
                    break
            annotation = getattr(annotation, '__origin__', None)
        else:
            if inspect.isclass(field) and issubclass(field, Field):
                # field is a Field subclass
                field = field()

        _cls = cls
        if not isinstance(field, Field):
            field_kwargs.update(
                default=default,
                immutable=final
            )
            field = cls.field_cls(**field_kwargs)
        else:
            _cls = field.parser_field_cls or _cls
            # this default is for the Annotated

        if not dependencies and field.dependencies:
            dependencies = field.dependencies

        if annotation is None:
            # a place to inject
            annotation = field.default_type

        input_type = _cls.rule_cls.parse_annotation(
            annotation=annotation,
            constraints=field.constraints,
            global_vars=global_vars,
            forward_refs=forward_refs,
            forward_key=attname,
            force_clear_refs=force_clear_refs
        )

        parser_field = _cls(
            attname=attname,
            name=(output_field or field).get_alias(
                attname, generator=options.alias_generator
            ),
            aliases=field.get_alias_from(
                attname, generator=options.alias_from_generator
            ),
            input_type=input_type,
            output_type=output_type,
            field=field,
            output_field=output_field,
            field_property=prop,
            dependencies=dependencies,
            # options=options,
            final=final,
            positional_only=positional_only,
            default=_default,
            **kwargs
        )
        parser_field.setup(options=options)
        return parser_field



import decimal
import uuid
from collections.abc import Mapping
from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import Union
from .base import TypeRegistry
import json
from .datastructures import unprovided
from ipaddress import IPv4Address, IPv6Address, IPv4Network, IPv6Network


encoder_registry = TypeRegistry('encoder', cache=True, shortcut='__encoder__')
register_encoder = encoder_registry.register


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        encoder = encoder_registry.resolve(type(o))
        if encoder:
            return encoder(o)
        return super().default(o)


class JSONSerializer:
    """
    Simple wrapper around json to be used in signing.dumps and
    signing.loads.
    """
    encoder_cls = JSONEncoder
    charset = 'utf-8'
    separators = (',', ':')
    ensure_ascii = False
    skipkeys = True

    def dumps(self, obj):
        return json.dumps(
            obj,
            separators=self.separators,
            cls=self.encoder_cls,
            ensure_ascii=self.ensure_ascii,
            skipkeys=self.skipkeys
        ).encode(self.charset)

    def loads(self, data: bytes):
        return json.loads(data.decode(self.charset))


def duration_iso_string(duration: timedelta):
    if duration < timedelta(0):
        sign = "-"
        duration *= -1
    else:
        sign = ""

    days = duration.days
    seconds = duration.seconds
    microseconds = duration.microseconds

    minutes = seconds // 60
    seconds = seconds % 60

    hours = minutes // 60
    minutes = minutes % 60

    ms = ".{:06d}".format(microseconds) if microseconds else ""
    return "{}P{}DT{:02d}H{:02d}M{:02d}{}S".format(
        sign, days, hours, minutes, seconds, ms
    )


@register_encoder(Mapping)
def from_mapping(data):
    return dict(data)


@register_encoder(set)
def from_set(data):
    return list(data)


@register_encoder(tuple)
def from_tuple(data):
    return list(data)


@register_encoder(unprovided.__class__)
def from_unprovided(data):
    return None


@register_encoder(bytes)
def from_bytes(data: bytes):
    return data.decode("utf-8", errors="replace")


@register_encoder(date)
def from_datetime(data: Union[datetime, date]):
    return data.isoformat()


@register_encoder(IPv4Network, IPv4Address, IPv6Network, IPv6Address)
def from_ip(data):
    return str(data)


@register_encoder(IPv4Network)
def from_datetime(data):
    return str(data)


@register_encoder(timedelta)
def from_duration(data: timedelta):
    return duration_iso_string(data)


@register_encoder(time)
def from_time(data: time):
    r = data.isoformat()
    if data.microsecond:
        r = r[:12]
    return r


@register_encoder(uuid.UUID)
def from_uuid(data: uuid.UUID):
    return str(data)


@register_encoder(decimal.Decimal)
def from_decimal(data: decimal.Decimal):
    if data.is_normal():
        if js_unsafe(data):
            return str(data)
        return float(data)
    return str(data)


# TODO?
# @register_encoder(float)
# def from_float(data: float):
#     if js_unsafe(data):
#         return str(data)
#     return data
#
#
# @register_encoder(int)
# def from_int(data: int):
#     if js_unsafe(data):
#         return str(data)
#     return data


@register_encoder(Enum)
def from_enum(en: Enum):
    return en.value


MAX_SAFE_NUMBER = 9007199254740991
MIN_SAFE_NUMBER = -9007199254740991


def js_unsafe(num: Union[int, float, decimal.Decimal]):
    return num > MAX_SAFE_NUMBER or num < MIN_SAFE_NUMBER


# @register_encoder(attr="__iter__")
# def from_iterable(encoder, data):
#     return list(data)
